from .primitive import PrimitiveBasis
from .legendre import LegendreBasis
from .fourier import FourierBasis
from .gsh import GSHBasis

__all__ = ['PrimitiveBasis', 'LegendreBasis', 'FourierBasis', 'GSHBasis']

DiscreteIndicatorBasis = PrimitiveBasis
ContinuousIndicatorBasis = PrimitiveBasis



"""Plotting functions for PyMKS notebooks
"""
import matplotlib.pyplot as plt
import numpy
from .func import curry, fmap


@curry
def _plot_ax(axis, arrs, titles, cmap, showticks):
    if hasattr(axis.get_subplotspec(), "colspan"):
        col_num = axis.get_subplotspec().colspan.start  # pragma: no cover
    else:
        col_num = axis.colNum  # pragma: no cover
    if not showticks:
        axis.set_xticks(())
        axis.set_yticks(())
    axis.set_title(titles[col_num])
    extent_ = lambda dim: (-arrs[col_num].shape[dim] / 2, arrs[col_num].shape[dim] / 2)
    return axis.imshow(
        arrs[col_num],
        interpolation="none",
        extent=extent_(1) + extent_(0),
        vmin=numpy.min(numpy.vstack(arrs)),
        vmax=numpy.max(numpy.vstack(arrs)),
        cmap=cmap,
    )


def _colorbar(fig, axis, image):
    """Generate a colorbar
    """
    axis.yaxis.set_offset_position("right")
    fig.colorbar(image, cax=axis)


def plot_microstructures(
    *arrs, titles=(), cmap=None, colorbar=True, showticks=False, figsize_weight=4
):
    """Plot a set of microstructures side-by-side

    Args:
      arrs: any number of 2D arrays to plot
      titles: a sequence of titles with len(*arrs)
      cmap: any matplotlib colormap

    >>> import numpy as np
    >>> np.random.seed(1)
    >>> x_data = np.random.random((2, 10, 10))
    >>> fig = plot_microstructures(
    ...     x_data[0],
    ...     x_data[1],
    ...     titles=['array 0', 'array 1'],
    ...     cmap='twilight'
    ... )
    >>> fig.show()  #doctest: +SKIP

    .. image:: plot_microstructures.png
       :width: 400


    """
    fig, axs = plt.subplots(
        1,
        len(arrs),
        figsize=(figsize_weight * len(arrs), figsize_weight),
        constrained_layout=True,
    )
    if len(arrs) == 1:
        axs = (axs,)
    if isinstance(titles, str):
        titles = (titles,)
    if len(titles) < len(arrs):
        titles = titles + ("",) * (len(arrs) - len(titles))
    plots = list(
        fmap(_plot_ax(arrs=arrs, titles=titles, cmap=cmap, showticks=showticks), axs)
    )
    if colorbar:
        _colorbar(
            fig, fig.add_axes([1.0, 0.05, 0.05, 0.9]), plots[0],
        )
    return fig



"""Test pair correlations
"""

import dask.array as da
import numpy as np
from pymks.fmks import correlations


def get_array():
    """Get array for tests
    """
    return da.from_array(
        np.array([[[1, 0, 0], [0, 1, 1], [1, 1, 0]], [[0, 0, 1], [1, 0, 0], [0, 0, 1]]])
    )


def get_mask():
    """Get mask for tests
    """
    mask = np.ones((2, 3, 3))
    mask[:, 2, 1:] = 0
    return da.from_array(mask)


def test_periodic():
    """Check that periodic still works

    The normalization occurs in the two_point_stats function and the
    auto-correlation/cross-correlation occur in the cross_correlation
    function. Checking that the normalization is properly calculated.
    First is the auto-correlation. Second is the cross-correlation.
    """
    array = get_array()
    correct = (
        (correlations.cross_correlation(array, array).compute() / 9)
        .round(3)
        .astype(np.float64)
    )
    tested = (
        correlations.two_point_stats(array, array).compute().round(3).astype(np.float64)
    )
    assert (correct == tested).all()

    correct = (
        (correlations.cross_correlation(array, 1 - array).compute() / 9)
        .round(3)
        .astype(np.float64)
    )
    tested = (
        correlations.two_point_stats(array, 1 - array)
        .compute()
        .round(3)
        .astype(np.float64)
    )
    assert (correct == tested).all()


def test_masked_periodic():
    """Check that masked periodic works

    Two point statistics are part correlation and part
    normalization. The correlation sums up the number of possible
    2-point states. In masked periodic, we assume that vectors going
    across the boundary of the structure come back on the other
    side. However, a vector landing in the masked area is discarded
    (ie not included in the correlation sum). Below, are the hand
    computed correlation and normalization. The correct 2point stats
    are the correlation divided by the normalization. First, is the
    auto-correlation and second is the cross-correlation.
    """

    correct_periodic_mask_auto = np.array(
        [[[2, 1, 2], [1, 4, 1], [2, 1, 2]], [[1, 0, 0], [0, 2, 0], [0, 0, 1]]]
    )

    correct_periodic_mask_cross = np.array(
        [[[1, 3, 1], [2, 0, 2], [1, 1, 1]], [[0, 1, 2], [2, 0, 2], [1, 2, 0]]]
    )

    norm_periodic_mask = np.array([[5, 5, 5], [6, 7, 6], [5, 5, 5]])

    mask = get_mask()
    array = get_array()

    # Auto-Correlation
    correct = (
        (correct_periodic_mask_auto / norm_periodic_mask).round(3).astype(np.float64)
    )
    tested = (
        correlations.two_point_stats(array, array, mask=mask, periodic_boundary=True)
        .compute()
        .round(3)
        .astype(np.float64)
    )

    assert (correct == tested).all()

    # Cross-Correlation
    correct = (
        (correct_periodic_mask_cross / norm_periodic_mask).round(3).astype(np.float64)
    )
    tested = (
        correlations.two_point_stats(
            array, 1 - array, mask=mask, periodic_boundary=True
        )
        .compute()
        .round(3)
        .astype(np.float64)
    )

    assert (correct == tested).all()


def test_non_periodic():
    """Test that non-periodic works

    Two point statistics are part correlation and part
    normalization. The correlation sums up the number of possible
    2-point states. In non-periodic, we assume that a vector used to
    count up 2 point states can only connect two states in the
    structure. A vector going outside of the bounds of the structure
    is not counted.

    Below, are the hand computed correlation and normalization. The
    correct 2point stats are the correlation divided by the
    normalization. First, is the auto-correlation and second is the
    cross-correlation.

    """
    correct_nonperiodic_auto = np.array(
        [[[1, 1, 2], [2, 5, 2], [2, 1, 1]], [[0, 0, 0], [0, 3, 0], [0, 0, 0]]]
    )

    correct_nonperiodic_cross = np.array(
        [[[2, 3, 1], [1, 0, 2], [0, 2, 1]], [[1, 2, 1], [2, 0, 1], [1, 2, 1]]]
    )

    norm_nonperiodic = np.array([[4, 6, 4], [6, 9, 6], [4, 6, 4]])

    array = get_array()
    # Auto-Correlation
    correct = (correct_nonperiodic_auto / norm_nonperiodic).round(3).astype(np.float64)
    tested = (
        correlations.two_point_stats(array, array, periodic_boundary=False)
        .compute()
        .round(3)
        .astype(np.float64)
    )

    assert (correct == tested).all()

    # Cross-Correlation
    correct = (correct_nonperiodic_cross / norm_nonperiodic).round(3).astype(np.float64)
    tested = (
        correlations.two_point_stats(array, 1 - array, periodic_boundary=False)
        .compute()
        .round(3)
        .astype(np.float64)
    )

    assert (correct == tested).all()


def test_non_periodic_masking():
    """Check that non-periodic masking works

    In non-periodic masking, vectors that go across the boundary or
    land in a mask are not included in the sum.

    """

    correct_nonperiodic_mask_auto = np.array(
        [[[1, 0, 1], [1, 4, 1], [1, 0, 1]], [[0, 0, 0], [0, 2, 0], [0, 0, 0]]]
    )

    correct_nonperiodic_mask_cross = np.array(
        [[[1, 3, 1], [1, 0, 1], [0, 1, 0]], [[0, 1, 1], [1, 0, 1], [1, 2, 0]]]
    )

    norm_nonperiodic_mask = np.array([[2, 4, 3], [4, 7, 4], [3, 4, 2]])

    array = get_array()
    mask = get_mask()

    # Auto-Correlation
    correct = (
        (correct_nonperiodic_mask_auto / norm_nonperiodic_mask)
        .round(3)
        .astype(np.float64)
    )
    tested = (
        correlations.two_point_stats(array, array, mask=mask, periodic_boundary=False)
        .compute()
        .round(3)
        .astype(np.float64)
    )
    assert (correct == tested).all()

    # Cross-Correlation
    correct = (
        (correct_nonperiodic_mask_cross / norm_nonperiodic_mask)
        .round(3)
        .astype(np.float64)
    )
    tested = (
        correlations.two_point_stats(
            array, 1 - array, mask=mask, periodic_boundary=False
        )
        .compute()
        .round(3)
        .astype(np.float64)
    )
    assert (correct == tested).all()


def test_different_sized_arrays():
    """Check that different sized dask arrays are valid masks.

    We want to be able to specify the same mask for each sample. We
    also want to be able to specify a different mask for each
    sample. This validates that both are possible.
    """

    array = da.random.random([1000, 3, 3])

    mask_same4all = da.random.randint(0, 2, [3, 3])
    mask_diff4all = da.random.randint(0, 2, [1000, 3, 3])

    correlations.two_point_stats(array, array, mask=mask_same4all)

    # The following check fails. Therefore, the current implementation
    # only works for one mask for all or different mask for all, which
    # is feature rich enough for me.
    # correlations.two_point_stats(array, array, mask=mask_same4some)
    correlations.two_point_stats(array, array, mask=mask_diff4all)


def test_bools_ints_valid():
    """Some check that boolean and integers are valid masks

    A mask could be true and false specifying where there is a
    microstructure. However, it could also be any value in the range
    $[0,1]$ which specifies the probability a value is correctly
    assigned. The mask right now only implements confidence in a
    single phase, although idealy it should represent the confidence
    in all phases. However, for the use cases where there are 2
    phases, a mask with a probability for one phase also completely
    describes the confidence in the other phase. Therefore, this
    implementation is complete for 2 phases.

    """

    mask_int = da.random.randint(0, 2, [1000, 3, 3])
    mask_bool = mask_int.copy().astype(bool)
    array = da.random.random([1000, 3, 3])

    correlations.two_point_stats(array, array, mask=mask_int)
    correlations.two_point_stats(array, array, mask=mask_bool)



from .base_microstructure_generator import _BaseMicrostructureGenerator
import numpy as np
from ..fmks.func import deprecate

class SphericalMicrostructureGenerator(_BaseMicrostructureGenerator):

    """
    Generates n_samples number of a periodic random spherical
    particles.  The particles have all the same phase and are in a
    matrix of a different phase (n_phase is always 2). The box is of
    size equal to size. The grain_size argument controls the size of
    the particles.

    In 1D.

    >>> size = (3, 3)
    >>> n_samples = 1
    >>> generator = SphericalMicrostructureGenerator(
    ...     n_samples=n_samples,
    ...     size=size,
    ...     n_particles=4,
    ...     grain_size=1.,
    ...     seed=10)
    >>> assert np.allclose(generator.generate(),
    ...                    [[[1, 0, 1], [1, 0, 1], [1, 0, 1]]])

    In 2D.

    >>> size = (3,)
    >>> n_samples = 1
    >>> generator = SphericalMicrostructureGenerator(
    ...     n_samples=n_samples,
    ...     size=size,
    ...     n_particles=2,
    ...     grain_size=2.,
    ...     seed=10)
    >>> assert np.allclose(generator.generate(), [[1, 1, 1]])

    In 3D.

    >>> size = (3, 3, 3)
    >>> n_samples = 1
    >>> generator = SphericalMicrostructureGenerator(
    ...     n_samples=n_samples,
    ...     size=size,
    ...     n_particles=2,
    ...     grain_size=2.,
    ...     seed=10)
    >>> X = [[[[0, 0, 0], [1, 1, 1], [0, 0, 0]],
    ...       [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
    ...       [[0, 0, 0], [1, 1, 1], [0, 0, 0]]]]
    >>> assert np.allclose(generator.generate(), X)

    """
    @deprecate
    def __init__(self, n_samples=1, size=(21, 21), n_particles=2,
                 grain_size=None, seed=3):
        """
        Instantiate a SphericalMicrostructureGenerator.

        Args:
          n_samples: number of samples to be generated
          size: size of samples
          n_particles: number of particles in the microstructures
          radius: radius of the particles
          seed: seed for random number generator

        Returns:
          n_samples number of a periodic random microstructure with size equal
          to size and with n_phases number of phases.
        """
        super(SphericalMicrostructureGenerator,
              self).__init__(n_samples=n_samples, size=size, n_phases=2,
                             grain_size=grain_size, seed=seed)
        self.n_particles = n_particles
        self.average_radius = self.grain_size / 2

    def _remove_overlaps(self, position, radius, size):
        r"""
        Remove points that overlap in a periodic box.

        >>> generator = SphericalMicrostructureGenerator()
        >>> size = np.array([4, 2])
        >>> position = np.array([[0.5, 0.5], [2., 0], [3., 1.]])
        >>> radius = np.array([0.5, 0.01, 0.5])
        >>> position_, radius_ = generator._remove_overlaps(position,
        ...                                                 radius, size)
        >>> assert np.allclose(position[1:], position_)
        >>> assert np.allclose(radius[1:], radius_)

        Args:
          position: array of points of shape (N, dim)
          radius: array of radii of shape (N,)
          size: size of box, array such that len(size) = dim

        Returns:
          return tuple of points and radii that do not overlap

        """

        dist_matrix = self._periodic_distance(
            position[:, None], position[None], size)
        radius_matrix = radius[None, :] + radius[:, None]
        mask_matrix = dist_matrix < radius_matrix
        mask_matrix = np.triu(mask_matrix, k=1)
        mask = np.any(mask_matrix, axis=1)
        return position[~mask], radius[~mask]

    def _periodic_distance(self, X, Y, size):
        r"""
        Calculate the periodic distance between points in a box. X and
        Y must be broadcastable.

        >>> generator = SphericalMicrostructureGenerator()
        >>> size = np.array([4, 2])
        >>> X = np.array([[0.5, 0.5], [2., 0]])
        >>> Y = np.array([[0., 0.], [1.5, 0.5], [3., 1.]])
        >>> r = np.sqrt(2.) / 2.
        >>> d = [[r, 1], [1, r], [r, 1]]
        >>> assert np.allclose(generator._periodic_distance(X[None],
        ...                                                 Y[:, None],
        ...                                                 size), d)

        Args:
          X: array of (..., dim)
          Y: array of (..., dim)
          size: size of box, array such that len(size) = dim

        Returns:
          Periodic distances between points in array X and Y, an array
          shaped according to the standard broadcast rules for X and Y.

        """
        a = (X - Y) % (size - 1)
        b = (Y - X) % (size - 1)
        d = np.minimum(a, b)
        return np.sqrt(np.sum(d**2, axis=-1))

    def _generate_sample(self):
        size = np.array(self.size)
        dim = len(size)
        particle_positions = np.random.random(
            (self.n_particles, dim)) * (size - 1)[None]
        radius = np.random.normal(
            self.average_radius, self.average_radius / 3., self.n_particles)
        radius = radius * np.sign(radius)
        particle_positions, radius = self._remove_overlaps(
            particle_positions, radius, size)
        xyz = (np.indices(size) + 0.5).swapaxes(0, -1)
        X_ = radius > self._periodic_distance(
            particle_positions[[None] * dim], xyz[..., None, :], size)
        return np.any(X_, axis=-1).astype(int)

    def generate(self):
        """
        Generates the microstructure.

        Returns:
          periodic microstructure
        """
        return np.array([self._generate_sample()
                         for _ in range(self.n_samples)])



from ..bases.imag_ffts import _ImagFFTBasis
import numpy as np
from ..fmks.func import deprecate

class _BaseMicrostructureGenerator(_ImagFFTBasis):
    @deprecate
    def __init__(self, n_samples=1, size=(21, 21),
                 n_phases=2, grain_size=None, seed=3, volume_fraction=None,
                 percent_variance=None):
        """
        Instantiate a MicrostructureGenerator.

        Args:
          n_samples (int): number of samples to be generated
          size (tuple): size of samples
          n_phases (int): number of phases in microstructures
          grain_size (tuple): size of the grain_size in the microstructure
          seed (int): seed for random number generator
          volume_fraction (tuple): the percent volume fraction for each phase.
          percent_variance (tuple): the percent variance for each value of
              volume_fraction. For example volume_fraction=(0.5, 0.5) and
              percent_variance=0.01 would be a 50 +/- 1 percent volume
              fractions for both values.

        Returns:
          n_samples number of a periodic random microstructure with size equal
          to size and with n_phases number of phases.
        """
        self._axes = np.arange(len(size)) + 1
        self._axes_shape = size
        self.n_samples = n_samples
        self.size = size
        self.n_phases = n_phases
        self.grain_size = grain_size
        if self.grain_size is None:
            self.grain_size = np.array(size) / 2
        np.random.seed(seed)
        self.volume_fraction = volume_fraction
        if self.volume_fraction is not None:
            if len(self.volume_fraction) != self.n_phases:
                raise RuntimeError(('n_phases and lenth of volume_fraction' +
                                   ' must be the same'))
            cum_frac = np.cumsum(volume_fraction)
            if not np.allclose(cum_frac[-1], 1):
                raise RuntimeError("volume fractions do not add up to 1")
            if percent_variance is None:
                percent_variance = 0.
            self.percent_variance = percent_variance
            min_frac = cum_frac[0] - percent_variance
            max_frac = cum_frac[-2] + percent_variance
            if max_frac > 1 or min_frac < 0:
                raise RuntimeError(('percent_variance cannot extend' +
                                    'volume_fraction values beyond 0 or 1'))
        super(_BaseMicrostructureGenerator, self).__init__()

    def generate(self):
        raise NotImplementedError



from .filter import Correlation
import numpy as np
from .fmks.func import deprecate
"""
The stats functions take in a microstructure function and returns its two
point statistics.
"""

@deprecate
def autocorrelate(X, basis, periodic_axes=[], n_jobs=1, confidence_index=None,
                  autocorrelations=None):
    """
    Computes the autocorrelation from a microstructure function.

    Args:
        X (ND array): The microstructure, an `(n_samples, n_x, ...)`
            shaped array where `n_samples` is the number of samples and
            `n_x` is the spatial discretization.
        basis (class): an instance of a bases class
        periodic_axes (list, optional): axes that are periodic. (0, 2) would
            indicate that axes x and z are periodic in a 3D microstrucure.
        n_jobs (int, optional): number of parallel jobs to run, only used if
            pyfftw is installed.
        confidence_index (ND array, optional): array with same shape as X used
            to assign a confidence value for each data point.
        autocorrelations (list, optional): list of spatial autocorrelations to
            be computed corresponding to the states in basis.n_states. For
            example, if basis.n_states=[0, 2], then autocorrelations=[(0, 0),
            (2, 2)] computes the autocorrelations for the states 0 and 2. If
            no list is passed, all autocorrelations in basis.n_states are
            computed.

    Returns:
        Autocorrelations for a microstructure.

    Non-periodic example

    >>> n_states = 2
    >>> X = np.array([[[0, 0, 0],
    ...                [0, 1, 0],
    ...                [0, 0, 0]]])
    >>> from pymks.bases import PrimitiveBasis
    >>> prim_basis = PrimitiveBasis(n_states=n_states)
    >>> X_auto = autocorrelate(X, prim_basis, periodic_axes=(0, 1))
    >>> X_test = np.array([[[0., 0., 0.],
    ...                   [0., 1./9, 0.],
    ...                   [0., 0., 0.]]])
    >>> assert(np.allclose(np.real_if_close(X_auto[0, ..., 1]), X_test[0]))
    """
    if periodic_axes is None:
        periodic_axes = []
    if autocorrelations is None:
        autocorrelations = _auto_correlations(basis.n_states)
    else:
        autocorrelations = _correlations_to_indices(autocorrelations, basis)
    return _compute_stats(X, basis, autocorrelations, confidence_index,
                          periodic_axes, n_jobs)

@deprecate
def crosscorrelate(X, basis, periodic_axes=None, n_jobs=1,
                   confidence_index=None, crosscorrelations=None):
    """
    Computes the crosscorrelations from a microstructure function.

    Args:
        X (ND array): The microstructure, an `(n_samples, n_x, ...)`
            shaped array where `n_samples` is the number of samples and
            `n_x` is the spatial discretization.s.
        basis (class): an instance of a bases class
        periodic_axes (list, optional): axes that are periodic. (0, 2) would
            indicate that axes x and z are periodic in a 3D microstrucure.
        n_jobs (int, optional): number of parallel jobs to run, only used if
            pyfftw is installed.
        confidence_index (ND array, optional): array with same shape as X used
            to assign a confidence value for each data point.
        crosscorrelations (list, optional): list of cross-correlations to
            be computed corresponding to the states in basis.n_states. For
            example if basis.n_states=[2, 4, 6] then crosscorrelations=[(2, 4),
            (2, 6)] computes the cross-correlations with local states 2 and 4
            as well as 2 and 6. If no list is passed, all cross-correlations
            in basis.n_states are computed.

    Returns:
        Crosscorelations for a microstructure.

    Examples

    Test for 2 states.

    >>> n_states = 2
    >>> X = np.array([[[0, 1, 0],
    ...                [0, 1, 0],
    ...                [0, 1, 0]]])
    >>> from pymks.bases import PrimitiveBasis
    >>> prim_basis = PrimitiveBasis(n_states=n_states)
    >>> X_cross = crosscorrelate(X, prim_basis, periodic_axes=[0, 1])
    >>> X_test = np.array([[[[1/3.], [0.], [1/3.]],
    ...                     [[1/3.], [0.], [1/3.]],
    ...                     [[1/3.], [0.], [1/3.]]]])
    >>> assert(np.allclose(X_cross, X_test))

    Test for 3 states

    >>> n_states = 3
    >>> prim_basis = PrimitiveBasis(n_states=n_states)
    >>> assert(crosscorrelate(X, prim_basis,
    ...        periodic_axes=[0, 1]).shape == (1, 3, 3, 3))

    Test for 4 states

    >>> n_states = 4
    >>> prim_basis = PrimitiveBasis(n_states=n_states)
    >>> assert(crosscorrelate(X, prim_basis,
    ...        periodic_axes=[0, 1]).shape == (1, 3, 3, 6))

    Test for 5 states

    >>> n_states = 5
    >>> prim_basis = PrimitiveBasis(n_states=n_states)
    >>> assert(crosscorrelate(X, prim_basis,
    ...        periodic_axes=[0, 1]).shape == (1, 3, 3, 10))
    """
    if periodic_axes is None:
        periodic_axes = []
    if crosscorrelations is None:
        crosscorrelations = _cross_correlations(basis.n_states)
    else:
        crosscorrelations = _correlations_to_indices(crosscorrelations,
                                                     basis)
    return _compute_stats(X, basis, crosscorrelations, confidence_index,
                          periodic_axes, n_jobs)

@deprecate
def correlate(X, basis, periodic_axes=None, n_jobs=1,
              confidence_index=None, correlations=None):
    """
    Computes the autocorrelations and crosscorrelations from a microstructure
    function.

    Args:
        X (ND array): The microstructure, an `(n_samples, n_x, ...)`
            shaped array where `n_samples` is the number of samples and
            `n_x` is the spatial discretization.
        basis (class): an instance of a bases class
        periodic_axes (list, optional): axes that are periodic. (0, 2) would
            indicate that axes x and z are periodic in a 3D microstrucure.
        n_jobs (int, optional): number of parallel jobs to run, only used if
            pyfftw is installed.
        confidence_index (ND array, optional): array with same shape as X used
            to assign a confidence value for each data point.
        correlations (list, optional): list of  spatial _check_shapes to
            be computed corresponding to the states in basis.n_states. For
            example, it n_states=[0, 2, 5] [(0, 0), (2, 2), (0, 5)] computes
            the autocorrelations with local states 0 and 2 as well as the
            cross-correlation between 0 and 5. If no list is passed, all
            spatial correlations are computed.

    Returns:
        Autocorrelations and crosscorrelations for a microstructure.

    Example

    >>> from pymks import PrimitiveBasis
    >>> prim_basis = PrimitiveBasis(2, [0, 1])
    >>>
    >>> np.random.seed(0)
    >>> X = np.random.randint(2, size=(1, 3))
    >>> X_corr = correlate(X, prim_basis)
    >>> X_result = np.array([[0, 0.5, 0],
    ...                      [1 / 3., 2 / 3., 0],
    ...                      [0, 0.5, 0.5]])
    >>> assert np.allclose(X_corr, X_result)
    """
    if periodic_axes is None:
        periodic_axes = []
    if correlations is None:
        L = basis.n_states
        _auto, _cross = _auto_correlations(L), _cross_correlations(L)
        correlations = (_auto[0] + _cross[0], _auto[1] + _cross[1])
    else:
        correlations = _correlations_to_indices(correlations, basis)
    return _compute_stats(X, basis, correlations, confidence_index,
                          periodic_axes, n_jobs)

@deprecate
def _compute_stats(X, basis, correlations, confidence_index,
                   periodic_axes, n_jobs):
    """Helper function to compute statistics

    Args:
        X (ND array): The microstructure, an `(n_samples, n_x, ...)`
            shaped array where `n_samples` is the number of samples and
            `n_x` is the spatial discretization.
        basis: an instance of a bases class
        correlations: list of  spatial correlations to be computed.
        confidence_index: array with same shape as X used to assign a
            confidence value for each data point.
        periodic_axes: axes that are periodic. (0, 2) would indicate that axes
            and z are periodic in a 3D microstrucure.
        n_jobs (int, optional): number of parallel jobs to run.
    """
    if max(max(correlations)) > len(basis.n_states) + 1:
        raise ValueError(('values in correlations are larger than') +
                         ('the length of basis.n_states'))
    X_ = basis.discretize(X)
    X_ = _mask_X_(X_, confidence_index)
    basis._n_jobs = n_jobs
    _Fkernel_shape(X.shape, basis, periodic_axes)
    _norm = _normalize(X_.shape, basis, confidence_index)
    return _correlate(X_, basis, correlations) / _norm

@deprecate
def _correlate(X_, basis, correlations):
    """
    Helper function used to calculate the unnormalized correlation counts.

    Args:
        X_ (ND array): The discretized microstructure function, an
            `(n_samples, n_x, ..., n_states)` shaped array
            where `n_samples` is the number of samples, `n_x` is thes
            patial discretization, and n_states is the number of local states.
        basis (class): an instance of a bases class.
        correlations (list): list of correlations to compute. `[(0, 0),
            (1, 1)]`

    Returns:
        correlation counts for a given microstructure function

    Example

    >>> from pymks.datasets import make_microstructure
    >>> from pymks.bases import PrimitiveBasis
    >>> X = make_microstructure(n_samples=2, n_phases=3,
    ...                         size=(2, 2), grain_size=(2, 2), seed=99)
    >>> prim_basis = PrimitiveBasis(n_states=3, domain=[0, 2])
    >>> X_ = prim_basis.discretize(X)
    >>> correlations = (tuple(prim_basis.n_states), tuple(prim_basis.n_states))
    >>> X_corr = _correlate(X_, prim_basis, correlations=correlations)
    >>> X_result = np.array([[[[0, 0, 0],
    ...                        [0, 0, 2]],
    ...                       [[0, 0, 0],
    ...                        [1, 1, 2]]],
    ...                      [[[0, 0, 0],
    ...                        [0, 0, 0]],
    ...                       [[2, 0, 2],
    ...                        [2, 0, 2]]]])
    >>> assert np.allclose(X_result, X_corr)
    """
    corr = Correlation(X_[..., correlations[0]],
                       basis).convolve(X_[..., correlations[1]])
    return _truncate(corr, X_.shape[:-1])

@deprecate
def _auto_correlations(n_states):
    """Returns list of autocorrelations

    Args:
        n_states: number of local states

    Returns:
        list of tuples for autocorrelations

    >>> l = _auto_correlations(np.arange(3))
    >>> assert l == ((0, 1, 2), (0, 1, 2))
    """
    return tuple(n_states), tuple(n_states)

@deprecate
def _cross_correlations(n_states):
    """Returns list of crosscorrelations

    Args:
        n_states: number of local states

    Returns:
        list of tuples for crosscorrelations

    >>> l = _cross_correlations(np.arange(3))
    >>> assert l == ((0, 0, 1), (1, 2, 2))
    """
    l = range(len(n_states))
    cross_corr = [[(l[i], l[j]) for j in l[1:][i:]] for i in l[:-1]]
    flat_corr = [item for sublist in cross_corr for item in sublist]
    l_0 = tuple([_l[0] for _l in flat_corr])
    l_1 = tuple([_l[1] for _l in flat_corr])
    return l_0, l_1

@deprecate
def _normalize(X_shape, basis, confidence_index):
    """
    Returns the normalization for the statistics

    The normalization should be Nx * Ny in the center of the domain.

    Args:
        `X_`: The discretized microstructure function, an
            `(n_samples, n_x, ..., n_states)` shaped array
            where `n_samples` is the number of samples, `n_x` is thes
            patial discretization, and n_states is the number of local states.
        basis (class): an instance of a bases class
        confidence_index (ND array, optional): array with same shape as X used
            to assign a confidence value for each data point.

    Returns:
        Normalization

    """

    if basis._axes_shape == X_shape[1:] and confidence_index is None:
        return float(np.prod(X_shape[1:]))
    else:
        mask = confidence_index
        if mask is None:
            mask = np.ones(X_shape[1:-1])[None]
        corr = Correlation(mask[..., None], basis)
        return _truncate(corr.convolve(mask[..., None]), X_shape)

@deprecate
def _Fkernel_shape(X_shape, basis, periodic_axes):
    """
    Assigns the shape of the kernel in Fourier space with non-periodic padding
    to the basis.

    Args:
        `X_shape`: The shape of discretized microstructure function,
            `(n_samples, n_x, ..., n_states)` where `n_samples` is the number
            of samples, `n_x` is the spatial discretization, and n_states is
            the number of local states.
        basis: an instance of a bases class
        periodic_axes: the axes of the array that are periodic

    Example

    >>> Nx = Ny = 5
    >>> X_ = np.zeros((1, Nx, Ny, 1))
    >>> periodic_axes = [1]
    >>> from pymks import PrimitiveBasis
    >>> p_basis = PrimitiveBasis(2)
    >>> p_basis._axes = np.array([1, 2])
    >>> _Fkernel_shape(X_.shape, p_basis, periodic_axes=periodic_axes)
    >>> assert p_basis._axes_shape == (10, 5)
    """
    a = np.ones(len(basis._axes), dtype=float) * 2
    a[list(periodic_axes)] = 1
    basis._axes_shape = tuple((np.array(X_shape)[basis._axes] * a).astype(int))

@deprecate
def _truncate(a, shape):
    """
    _truncates the edges of the array, a, based on the shape. This is
    used to unpad a padded convolution.

    Args:
        a: array to be truncated
        shape: new shape of array

    Returns:
        truncated array

    Example

    >>> print(_truncate(np.arange(10).reshape(1, 10, 1), (1, 5))[0, ..., 0])
    [3 4 5 6 7]
    >>> print(_truncate(np.arange(9).reshape(1, 9, 1), (1, 5))[0, ..., 0])
    [2 3 4 5 6]
    >>> print(_truncate(np.arange(10).reshape((1, 10, 1)), (1, 4))[0, ..., 0])
    [3 4 5 6]
    >>> print(_truncate(np.arange(9).reshape((1, 9, 1)), (1, 4))[0, ..., 0])
    [2 3 4 5]

    >>> a = np.arange(5 * 4).reshape((1, 5, 4, 1))
    >>> print(_truncate(a, shape=(1, 3, 2))[0, ..., 0])
    [[ 5  6]
     [ 9 10]
     [13 14]]

    >>> a = np.arange(5 * 4 * 3).reshape((1, 5, 4, 3, 1))
    >>> assert (_truncate(a, (1, 2, 2, 1))[0, ..., 0]  ==
    ...         [[[16], [19]], [[28], [31]]]).all()

    """
    a_shape = np.array(a.shape)
    n = len(shape)
    new_shape = a_shape.copy()
    new_shape[:n] = shape
    diff_shape = a_shape - new_shape
    index0 = (diff_shape + (diff_shape % 2) * (new_shape % 2)) // 2
    index1 = index0 + new_shape
    multi_slice = tuple(slice(index0[ii], index1[ii]) for ii in range(n))
    return a[multi_slice]

@deprecate
def _mask_X_(X_, confidence_index):
    """
    Helper function to verify that the confidence_index is the correct
    shape.

    Args:
        `X_`: The discretized microstructure function, an
            `(n_samples, n_x, ..., n_states)` shaped array
            where `n_samples` is the number of samples, `n_x` is thes
            patial discretization, and n_states is the number of local states.
        confidence_index: array with same shape as X used to assign a
            confidence value for each data point.

    Returns:
        The discretized microstructure function scaled by the confidence_index.

    """
    if confidence_index is not None:
        if X_.shape[:-1] != confidence_index.shape:
            raise RuntimeError('confidence_index does not match shape of X')
        X_ = X_ * confidence_index[..., None]
    return X_

@deprecate
def _correlations_to_indices(correlations, basis):
    """
    Helper function to select correct indices given the local state values in
    basis.n_states.

    Args:
        correlations: list of correlations to be computed
        basis: an instance of a basis class.

    Returns:
        list of correlations in terms of indices
    """
    try:
        l_0 = tuple([list(basis.n_states).index(_l[0]) for _l in correlations])
        l_1 = tuple([list(basis.n_states).index(_l[1]) for _l in correlations])
    except ValueError as ve:
        raise ValueError('correlations value ' + str(ve) +
                         ' is not in basis.n_states')
    return (l_0, l_1)



"""Test functions for graph descriptors
"""

import pytest
import numpy as np

GRASPI = pytest.importorskip("graspi")


def test_compute_descriptors():
    """Test the main cython function in Graspi.
    """
    morph = np.array([1, 1, 0, 0, 0, 1, 1, 0, 0, 1, 1, 0], dtype=np.int32)
    expected = [
        (12.0, b"STAT_n"),
        (8.0, b"STAT_e"),
        (6.0, b"STAT_n_D"),
        (6.0, b"STAT_n_A"),
        (1.0, b"STAT_CC_D"),
        (1.0, b"STAT_CC_A"),
        (1.0, b"STAT_CC_D_An"),
        (1.0, b"STAT_CC_A_Ca"),
        (0.4804587662220001, b"ABS_wf_D"),
        (0.5, b"ABS_f_D"),
        (0.8879592418670654, b"DISS_wf10_D"),
        (1.0, b"DISS_f10_D"),
        (1.0, b"CT_f_e_conn"),
        (1.0, b"CT_f_conn_D_An"),
        (1.0, b"CT_f_conn_A_Ca"),
        (8.0, b"CT_e_conn"),
        (6.0, b"CT_e_D_An"),
        (6.0, b"CT_e_A_Ca"),
        (1.0, b"CT_f_D_tort1"),
        (0.8333333134651184, b"CT_f_A_tort1"),
        (3.0, b"CT_n_D_adj_An"),
        (3.0, b"CT_n_A_adj_Ca"),
    ]

    assert GRASPI.compute_descriptors(morph, 4, 3, 1, 2, 1) == expected



from pymks.datasets import make_elastic_FE_strain_random
from pymks.datasets import make_elastic_FE_strain_delta
from pymks.datasets import make_elastic_stress_random
import numpy as np


def test_make_elastic_FE_strain_delta():
    elastic_modulus = (1., 2.)
    poissons_ratio = (0.3, 0.3)
    X, y = make_elastic_FE_strain_delta(elastic_modulus=elastic_modulus,
                                        poissons_ratio=poissons_ratio,
                                        size=(5, 5))


def test_make_elastic_FE_strain_random():
    elastic_modulus = (1., 2.)
    poissons_ratio = (0.3, 0.3)
    X, y = make_elastic_FE_strain_random(n_samples=1,
                                         elastic_modulus=elastic_modulus,
                                         poissons_ratio=poissons_ratio,
                                         size=(5, 5))


def test_make_elastic_stress_randome():
    X, y = make_elastic_stress_random(n_samples=1, elastic_modulus=(1, 1),
                                      poissons_ratio=(1, 1),
                                      grain_size=(3, 3), macro_strain=1.0)
    assert np.allclose(y, np.ones(y.shape))
    X, y = make_elastic_stress_random(n_samples=1, grain_size=(1, 1),
                                      elastic_modulus=(100, 200),
                                      size=(2, 2), poissons_ratio=(1, 3),
                                      macro_strain=1., seed=4)
    X_result = np.array([[[1, 1],
                          [0, 1]]])
    assert float(np.round(y, decimals=5)[0]) == 228.74696
    assert np.allclose(X, X_result)
    X, y = make_elastic_stress_random(n_samples=1, grain_size=(1, 1, 1),
                                      elastic_modulus=(100, 200),
                                      poissons_ratio=(1, 3),  seed=5,
                                      macro_strain=1., size=(2, 2, 2))
    X_result = np.array([[[1, 1],
                          [1, 0]],
                         [[1, 1],
                          [0, 0]]])
    assert np.allclose(X, X_result)
    assert y.astype(int) == 145



from pymks.stats import correlate
from sklearn.base import BaseEstimator
from sklearn.decomposition import PCA
import numpy as np
from .fmks.func import deprecate

class MKSStructureAnalysis(BaseEstimator):
    """
    `MKSStructureAnalysis` computes the 2-point statistics for a set of
    microstructures and does dimensionality reduction. It can be used to
    evaluate the selection of spatial correlations and look at clustering of
    2-point statistics.

    Attributes:
        n_components: Number of components used by `dimension_reducer`.
        dimension_reducer: Instance of a dimensionality reduction class.
        correlations: spatial correlations to be computed
        basis: instance of a basis class
        reduced_fit_data: Low dimensionality representation of spatial
            correlations used to fit the components.
        reduced_transformed_data: Reduced of spatial correlations.
        periodic_axes: axes that are periodic. (0, 2) would indicate that
            axes x and z are periodic in a 3D microstrucure.
        transformed_correlations: spatial correlations transform into the Low
            dimensional space.



    Below is an example of using MKSStructureAnalysis using FastICA.

    >>> from pymks.datasets import make_microstructure
    >>> from pymks.bases import PrimitiveBasis
    >>> from sklearn.decomposition import FastICA

    >>> leg_basis = PrimitiveBasis(n_states=2, domain=[0, 1])
    >>> reducer = FastICA(n_components=3)
    >>> analyzer = MKSStructureAnalysis(basis=leg_basis, mean_center=False,
    ...                                 dimension_reducer=reducer)

    >>> X = make_microstructure(n_samples=4, size=(13, 13), grain_size=(3, 3))
    >>> print(analyzer.fit_transform(X)) # doctest: +ELLIPSIS
    [[ 0.5 -0.5 -0.5]
     [ 0.5  0.5  0.5]
     [-0.5 -0.5  0.5]
     [-0.5  0.5 -0.5]]

    """
    @deprecate
    def __init__(self, basis, correlations=None, dimension_reducer=None,
                 n_components=None, periodic_axes=None,
                 store_correlations=False, n_jobs=1, mean_center=True):
        """
        Create an instance of a `MKSStructureAnalysis`.

        Args:
            basis: an instance of a bases class.
            dimension_reducer (class, optional): an instance of a
                dimensionality reduction class with a fit_transform method. The
                default class is PCA.
            n_components (int, optional): number of components kept by the
                dimension_reducer
            correlations (list, optional): list of spatial correlations to
                compute, default is the autocorrelation with the first local
                state and all of its cross correlations. For example if basis
                has basis.n_states=3, correlation would be [(0, 0), (0, 1),
                (0, 2)]. If n_states=[0, 2, 4], the default correlations are
                [(0, 0), (0, 2), (0, 4)] corresponding to the autocorrelations
                for the 0th local state, and the cross correlations with the 0
                and 2 as well as 0 and 4.
            periodic_axes (list, optional): axes that are periodic. (0, 2)
                would indicate that axes x and z are periodic in a 3D
                microstrucure.
            store_correlations (boolean, optional): If true the computed
                2-point statistics will be saved as an attributes
                fit_correlations and transform_correlations.
            n_jobs (int, optional): number of parallel jobs to run, only used
                if pyfftw is installed.
            mean_center (boolean, optional): If true the data will be mean
                centered before dimensionality reduction is computed.
        """
        self.basis = basis
        self.correlations = correlations
        self.dimension_reducer = dimension_reducer
        self.store_correlations = store_correlations
        self.mean_center = mean_center
        self.periodic_axes = periodic_axes
        if basis is not None:
            self.basis._n_jobs = n_jobs
        if self.dimension_reducer is None:
            self.dimension_reducer = PCA(svd_solver='randomized', copy=False)
        if n_components is None:
            n_components = self.dimension_reducer.n_components
        if n_components is None:
            n_components = 5
        self.n_components = n_components
        if self.correlations is None and basis is not None:
            correlations = [(0, l) for l in range(len(self.basis.n_states))]
            self.correlations = correlations
        if not callable(getattr(self.dimension_reducer,
                                "fit_transform", None)):
            raise RuntimeError(
                "dimension_reducer does not have fit_transform() method.")
        if not callable(getattr(self.dimension_reducer, "transform", None)):
            raise RuntimeError(
                "dimension_reducer does not have transform() method.")

    @property
    def n_components(self):
        return self._n_components

    @n_components.setter
    def n_components(self, value):
        """Setter for the number of components used by the dimension_reducer
        """
        self._n_components = value
        self.dimension_reducer.n_components = value

    @property
    def components_(self):
        stats_shape = ((self.n_components,) + self._components_shape)
        return self.dimension_reducer.components_.reshape(stats_shape)

    @components_.setter
    def components_(self, components):
        """Setter for the components used by the dimension_reducer
        """
        self.dimension_reducer.components_ = components.reshape(
            self._n_components, -1)

    def fit(self, X, reducer_labels=None, confidence_index=None):
        """Fits data by using the 2-point statistics for X to fits the
        components used in dimensionality reduction.

        Args:
            X (ND array): The microstructures or spatial correlations, a
                `(n_samples, n_x, ...)` shaped array where `n_samples` is the
                number of samples and `n_x` is the spatial discretization.
            reducer_labels (1D array, optional): label for X used during the
                fit_transform method for the `dimension_reducer`.
            confidence_index (ND array, optional): array with same shape as X
                used to assign a confidence value for each data point.

        Example

        >>> from pymks.datasets import make_delta_microstructures
        >>> from pymks import PrimitiveBasis
        >>> n_states = 2
        >>> analyzer = MKSStructureAnalysis(basis=PrimitiveBasis(n_states),
        ...                              n_components=1)
        >>> np.random.seed(5)
        >>> size = (2, 3, 3)
        >>> X = np.random.randint(2, size=size)
        >>> analyzer.fit(X)

        >>> res = analyzer.dimension_reducer.components_.reshape(size)[0]
        >>> sol = [[0.02886463, 0.02886463, 0.02886463],
        ...        [0.02886463, -0.43874233, 0.49647159],
        ...        [0.02886463, 0.02886463, -0.17896069]]

        >>> assert np.allclose(res, sol) or np.allclose(-res, sol)

        """
        X_stats = self._compute_stats(X, confidence_index)
        self._fit_transform(X_stats, reducer_labels)

    def fit_transform(self, X, confidence_index=None):
        """Fits data by using the 2-point statistics for X to fits the
        components used in dimensionality reduction and returns the reduction
        of the 2-point statistics for X.

        Args:
            X (ND array): The microstructures or spatial correlations, a
                `(n_samples, n_x, ...)` shaped array where `n_samples` is the
                number of samples and `n_x` is the spatial discretization.
            reducer_labels (1D array, optional): label for X used during the
                fit_transform method for the `dimension_reducer`..
            confidence_index (ND array, optional): array with same shape as X
                used to assign a confidence value for each data point.

        Returns:
           Reduction of the 2-point statistics of X used to fit the components.

        Example

        >>> from pymks.datasets import make_delta_microstructures
        >>> from pymks import PrimitiveBasis
        >>> n_states = 2
        >>> analyzer = MKSStructureAnalysis(basis=PrimitiveBasis(n_states),
        ...                              n_components=1)
        >>> np.random.seed(5)
        >>> size = (2, 3, 3)
        >>> X = np.random.randint(2, size=size)
        >>>
        >>> res = analyzer.fit_transform(X)
        >>> sol = [[ 0.26731852], [-0.26731852]]
        >>> assert np.allclose(res, sol) or np.allclose(-res, sol)

        """
        X_stats = self._compute_stats(X, confidence_index)
        return self._fit_transform(X_stats, None)

    def transform(self, X, confidence_index=None):
        """Computes the 2-point statistics for X and applies dimensionality
        reduction.

        Args:
            X (ND array): The microstructures or spatial correlations, a
                `(n_samples, n_x, ...)` shaped array where `n_samples` is the
                number of samples and `n_x` is the spatial discretization.
            confidence_index (ND array, optional): array with same shape as X
                used to assign a confidence value for each data point.

        Returns:
           Reduction of the 2-point statistics of X.

        Example

        >>> from pymks.datasets import make_delta_microstructures
        >>> from pymks import PrimitiveBasis
        >>> n_states = 2
        >>> analyzer = MKSStructureAnalysis(basis=PrimitiveBasis(n_states),
        ...                              n_components=1)
        >>> np.random.seed(5)
        >>> size = (2, 3, 3)
        >>> X = np.random.randint(2, size=size)

        >>> res = analyzer.fit_transform(X)
        >>> sol = [[ 0.26731852], [-0.26731852]]
        >>> assert np.allclose(res, sol) or np.allclose(-res, sol)

        >>> res = analyzer.fit_transform(X)
        >>> sol = [[ 0.26731852], [-0.26731852]]
        >>> assert np.allclose(res, sol) or np.allclose(-res, sol)

        """
        X_stats = self._compute_stats(X, confidence_index)
        return self._transform(X_stats)

    def _transform(self, X):
        """Reshapes and reduces X"""
        self._store_correlations(X)
        X_reshaped = self._reduce_shape(X)
        self.transform_data = self.dimension_reducer.transform(X_reshaped)
        return self.transform_data

    def _fit_transform(self, X, y):
        """Reshapes X and uses it to compute the components"""
        if self.store_correlations:
            self.fit_correlations = X.copy()
        X_reshaped = self._reduce_shape(X)
        self.reduced_fit_data = self.dimension_reducer.fit_transform(
            X_reshaped, y)
        self._components_shape = X.shape[1:-2] + (X.shape[-1] * X.shape[-2],)
        return self.reduced_fit_data

    def _store_correlations(self, X):
        """store stats"""
        if self.store_correlations:
            if hasattr(self, 'transform_correlations'):
                self.transform_correlations = np.concatenate(
                    (self.transform_correlations, X.copy()))
            else:
                self.transform_correlations = X.copy()

    def _compute_stats(self, X, confidence_index):
        """
        Helper function used to calculated 2-point statistics from `X` and
        reshape them appropriately for fit and predict methods.

        Args:
            X (ND array): The microstructure, an `(n_samples, n_x, ...)`
                shaped array where `n_samples` is the number of samples and
                `n_x` is the spatial discretization..
            confidence_index (ND array, optional): array with same shape as X
                used to assign a confidence value for each data point.

        Returns:
            Spatial correlations for each sample formated with dimensions
            (n_samples, n_features).

        Example
        """
        if self.basis is None:
            raise AttributeError('basis must be specified')
        X_stats = correlate(X, self.basis, periodic_axes=self.periodic_axes,
                            confidence_index=confidence_index,
                            correlations=self.correlations)
        return X_stats

    def _reduce_shape(self, X_stats):
        """
        Helper function used to reshape 2-point statistics appropriately for
        fit and predict methods.

        Args:
            `X_stats`: The discretized microstructure function, an
                `(n_samples, n_x, ..., n_states)` shaped array
                Where `n_samples` is the number of samples, `n_x` is the
                spatial discretization, and n_states is the number of local
                states.

        Returns:
            Spatial correlations for each sample formated with dimensions
            (n_samples, n_features).

        """
        X_reshaped = X_stats.reshape((X_stats.shape[0], X_stats[0].size))
        if self.mean_center:
            X_reshaped -= np.mean(X_reshaped,
                                  axis=1)[:, None].astype(X_reshaped.dtype)
        return X_reshaped



from .cahn_hilliard_simulation import CahnHilliardSimulation
from .microstructure_generator import MicrostructureGenerator
from pymks import DiscreteIndicatorBasis, MKSRegressionModel
import numpy as np
from ..fmks.func import deprecate

__all__ = ['make_delta_microstructures', 'make_elastic_FE_strain_delta',
           'make_elastic_FE_strain_random', 'make_cahn_hilliard',
           'make_microstructure', 'make_checkerboard_microstructure',
           'make_elastic_stress_random', 'make_elastic_stiffness']


@deprecate
def make_elastic_FE_strain_delta(elastic_modulus=(100, 150),
                                 poissons_ratio=(0.3, 0.3),
                                 size=(21, 21), macro_strain=0.01):
    """Generate delta microstructures and responses

    Simple interface to generate delta microstructures and their
    strain response fields that can be used for the fit method in the
    `MKSRegressionModel`. The length of `elastic_modulus` and
    `poissons_ratio` indicates the number of phases in the
    microstructure. The following example is or a two phase
    microstructure with dimensions of `(5, 5)`.

    Args:
        elastic_modulus (list, optional): elastic moduli for the phases
        poissons_ratio (list, optional): Poisson's ratios for the phases
        size (tuple, optional): size of the microstructure
        macro_strain (float, optional): Scalar for macroscopic strain applied
        strain_index (int, optional): interger value to return a particular
            strain field. 0 returns exx, 1 returns eyy, etc. To return all
            strain fields set strain_index equal to slice(None).

    Returns:
        tuple containing delta microstructures and their strain fields

    """
    from pymks.fmks.data.elastic_fe import solve

    X = make_delta_microstructures(len(elastic_modulus), size=size)

    return X, solve(X,
                    elastic_modulus=elastic_modulus,
                    poissons_ratio=poissons_ratio,
                    macro_strain=macro_strain)['strain'][..., 0]

@deprecate
def make_delta_microstructures(n_phases=2, size=(21, 21)):
    """Constructs delta microstructures

    Constructs delta microstructures for an arbitrary number of phases
    given the size of the domain.

    Args:
        n_phases (int, optional): number of phases
        size (tuple, optional): dimension of microstructure

    Returns:
        delta microstructures for the system of shape
        (n_samples, n_x, ...)

    Example

    >>> X = np.array([[[[0, 0, 0],
    ...                 [0, 0, 0],
    ...                 [0, 0, 0]],
    ...                [[0, 0, 0],
    ...                 [0, 1, 0],
    ...                 [0, 0, 0]],
    ...                [[0, 0, 0],
    ...                 [0, 0, 0],
    ...                 [0, 0, 0]]],
    ...               [[[1, 1, 1],
    ...                 [1, 1, 1],
    ...                 [1, 1, 1]],
    ...                [[1, 1, 1],
    ...                 [1, 0, 1],
    ...                 [1, 1, 1]],
    ...                [[1, 1, 1],
    ...                 [1, 1, 1],
    ...                 [1, 1, 1]]]])

    >>> assert(np.allclose(X, make_delta_microstructures(2, size=(3, 3, 3))))

    """
    shape = (n_phases, n_phases) + size
    center = tuple((np.array(size) - 1) // 2)
    X = np.zeros(shape=shape, dtype=int)
    X[:] = np.arange(n_phases)[(slice(None), None) + (None,) * len(size)]
    X[(slice(None), slice(None)) + center] = np.arange(n_phases)
    mask = ~np.identity(n_phases, dtype=bool)
    return X[mask]

@deprecate
def make_elastic_FE_strain_random(n_samples=1, elastic_modulus=(100, 150),
                                  poissons_ratio=(0.3, 0.3), size=(21, 21),
                                  macro_strain=0.01):
    """Generate random microstructures and responses

    Simple interface to generate random microstructures and their
    strain response fields that can be used for the fit method in the
    `MKSRegressionModel`. The following example is or a two phase
    microstructure with dimensions of `(5, 5)`.

    Args:
        n_samples (int, optional): number of samples
        elastic_modulus (list, optional): elastic moduli for the phases
        poissons_ratio (list, optional): Poisson's ratios for the phases
        size (tuple, optional): size of the microstructure
        macro_strain (float, optional): Scalar for macroscopic strain applied
        strain_index (int, optional): interger value to return a particular
            strain field. 0 returns exx, 1 returns eyy, etc. To return all
            strain fields set strain_index equal to slice(None).

    Returns:
         tuple containing delta microstructures and their strain fields

    """

    from pymks.fmks.data.elastic_fe import solve

    X = np.random.randint(len(elastic_modulus), size=((n_samples, ) + size))

    return X, solve(X,
                    elastic_modulus=elastic_modulus,
                    poissons_ratio=poissons_ratio,
                    macro_strain=macro_strain)['strain'][..., 0]

@deprecate
def make_cahn_hilliard(n_samples=1, size=(21, 21), dx=0.25, width=1.,
                       dt=0.001, n_steps=1):
    """Generate microstructures and responses for Cahn-Hilliard.
    Simple interface to generate random concentration fields and their
    evolution after one time step that can be used for the fit method in the
    `MKSRegressionModel`.  The following example is or a two phase
    microstructure with dimensions of `(6, 6)`.

    Args:
        n_samples (int, optional): number of microstructure samples
        size (tuple, optional): size of the microstructure
        dx (float, optional): grid spacing
        dt (float, optional): timpe step size
        width (float, optional): interface width between phases.
        n_steps (int, optional): number of time steps used

    Returns:
        Array representing the microstructures at n_steps ahead of 'X'

    Example

    >>> X, y = make_cahn_hilliard(n_samples=1, size=(6, 6))

    `X` is the initial concentration fields, and `y` is the
    strain response fields (the concentration after one time step).

    """
    CHsim = CahnHilliardSimulation(dx=dx, dt=dt, gamma=width ** 2)

    X0 = 2 * np.random.random((n_samples,) + size) - 1
    X = X0.copy()
    for ii in range(n_steps):
        CHsim.run(X)
        X = CHsim.response
    return X0, X

@deprecate
def make_microstructure(n_samples=10, size=(101, 101), n_phases=2,
                        grain_size=None, seed=10, volume_fraction=None,
                        percent_variance=None):
    """
    Constructs microstructures for an arbitrary number of phases
    given the size of the domain, and relative grain size.

    Args:
        n_samples (int, optional): number of samples
        size (tuple, optional): dimension of microstructure
        n_phases (int, optional): number of phases
        grain_size (tuple, optional): effective dimensions of grains
        seed (int, optional): seed for random number microstructureGenerator
        volume_fraction(tuple, optional): specify the volume fraction of each
            phase
        percent_variance(int, optional): varies the volume fraction of the
            microstructure up to this percentage

    Returns:
        microstructures for the system of shape (n_samples, n_x, ...)

    Example

    >>> n_samples, n_phases = 1, 2
    >>> size, grain_size = (3, 3), (1, 1)
    >>> Xtest = np.array([[[0, 1, 0],
    ...                [0, 0, 0],
    ...                [0, 1, 1]]])
    >>> X = make_microstructure(n_samples=n_samples, size=size,
    ...                         n_phases=n_phases, grain_size=grain_size,
    ...                         seed=0)

    >>> assert(np.allclose(X, Xtest))

    """
    if grain_size is None:
        grain_size = np.array(size) / 3.
    if np.sum(np.array(grain_size) > np.array(size)):
        raise RuntimeError('grain_size must be smaller than size')
    MS = MicrostructureGenerator(n_samples=n_samples, size=size,
                                 n_phases=n_phases, grain_size=grain_size,
                                 seed=seed, volume_fraction=volume_fraction,
                                 percent_variance=percent_variance)
    return MS.generate()

@deprecate
def make_checkerboard_microstructure(square_size, n_squares):
    """
    Constructs a checkerboard_microstructure with the `square_size` by
    `square_size` size squares and on a `n_squares` by `n_squares`

    Args:
        square_size (int): length of the side of one square in the
            checkerboard.
        n_squares (int): number of squares along on size of the checkerboard.

    Returns:
        checkerboard microstructure with shape of (1, square_size * n_squares,
        square_size * n_squares)

    Example

    >>> square_size, n_squares = 2, 2
    >>> Xtest = np.array([[[0, 0, 1, 1],
    ...                    [0, 0, 1, 1],
    ...                    [1, 1, 0, 0],
    ...                    [1, 1, 0, 0]]])
    >>> X = make_checkerboard_microstructure(square_size, n_squares)
    >>> assert(np.allclose(X, Xtest))

    """

    L = n_squares * square_size
    X = np.ones((2 * square_size, 2 * square_size), dtype=int)
    X[:square_size, :square_size] = 0
    X[square_size:, square_size:] = 0
    return np.tile(X, (int((n_squares + 1) / 2),
                   int((n_squares + 1) / 2)))[None, :L, :L]

@deprecate
def make_elastic_stress_random(n_samples=[10, 10], elastic_modulus=(100, 150),
                               poissons_ratio=(0.3, 0.3), size=(21, 21),
                               macro_strain=0.01, grain_size=[(3, 3), (9, 9)],
                               seed=10, volume_fraction=None,
                               percent_variance=None):
    """
    Generates microstructures and their macroscopic stress values for an
    applied macroscopic strain.

    Args:
        n_samples (int, optional): number of samples
        elastic_modulus (tuple, optional): list of elastic moduli for the
            different phases.
        poissons_ratio (tuple, optional): list of poisson's ratio values for
            the phases.
        size (tuple, optional): size of the microstructures
        macro_strain (tuple, optional): macroscopic strain applied to the
            sample.
        grain_size (tuple, optional): effective dimensions of grains
        seed (int, optional): seed for random number generator
        volume_fraction(tuple, optional): specify the volume fraction of
            each phase
        percent_variance(int, optional): Only used if volume_fraction is
            specified. Randomly varies the volume fraction of the
            microstructure.


    Returns:
        array of microstructures with dimensions (n_samples, n_x, ...) and
        effective stress values

    """
    if not isinstance(grain_size[0], (list, tuple, np.ndarray)):
        grain_size = (grain_size,)
    if not isinstance(n_samples, (list, tuple, np.ndarray)):
        n_samples = (n_samples,)
    if volume_fraction is None:
        volume_fraction = (None,) * len(n_samples)
    vf_0 = volume_fraction[0]
    if not isinstance(vf_0, (list, tuple, np.ndarray)) and vf_0 is not None:
        volume_fraction = (volume_fraction,)
        np.random.seed(seed)
    if not isinstance(size, (list, tuple, np.ndarray)) or len(size) > 3:
        raise RuntimeError('size must have length of 2 or 3')
    [RuntimeError('dimensions of size and grain_size are not the same.')
     for grains in grain_size if len(size) != len(grains)]
    if vf_0 is not None:
        [RuntimeError('dimensions of size and grain_size are not the same.')
         for volume_frac in volume_fraction
         if len(elastic_modulus) != len(volume_frac)]
    if len(elastic_modulus) != len(poissons_ratio):
        raise RuntimeError('length of elastic_modulus and poissons_ratio are' \
                           'not the same.')
    np.random.seed(seed)
    seed = np.random.randint(100, size=(len(volume_fraction),))
    X_cal, y_cal = make_elastic_FE_strain_delta(elastic_modulus,
                                                poissons_ratio, size,
                                                macro_strain)
    n_states = len(elastic_modulus)
    basis = DiscreteIndicatorBasis(n_states)
    model = MKSRegressionModel(basis=basis)
    model.fit(X_cal, y_cal)
    X = np.concatenate([make_microstructure(n_samples=sample, size=size,
                                            n_phases=n_states,
                                            grain_size=gs, seed=s,
                                            volume_fraction=vf,
                                            percent_variance=percent_variance)
                        for vf, s, gs, sample in zip(volume_fraction, seed,
                                                  grain_size, n_samples)])
    X_ = basis.discretize(X)
    index = tuple([None for i in range(len(size) + 1)]) + (slice(None),)
    modulus = np.sum(X_ * np.array(elastic_modulus)[index], axis=-1)
    y_stress = model.predict(X) * modulus
    return X, np.average(y_stress.reshape(len(y_stress), -1), axis=1)

@deprecate
def make_elastic_stiffness(n_samples=[10, 10],
                           elastic_modulus=(100, 150),
                           poissons_ratio=(0.3, 0.3), size=(21, 21),
                           macro_strain=0.01,
                           grain_size=[(3, 3), (9, 9)],
                           seed=10, volume_fraction=None,
                           percent_variance=None):
    """
    Generates microstructures and their effective stiffness values for an
    applied macroscopic strain.

    Args:
        n_samples (int, optional): number of samples
        elastic_modulus (tuple, optional): list of elastic moduli for the
            different phases.
        poissons_ratio (tuple, optional): list of poisson's ratio values for
            the phases.
        size (tuple, optional): size of the microstructures
        macro_strain (tuple, optional): macroscopic strain applied to the
            sample.
        grain_size (tuple, optional): effective dimensions of grains
        seed (int, optional): seed for random number generator
        volume_fraction(tuple, optional): specify the volume fraction of
            each phase
        percent_variance(int, optional): Only used if volume_fraction is
            specified. Randomly varies the volume fraction of the
            microstructure.


    Returns:
        array of microstructures with dimensions (n_samples, n_x, ...) and
        effective stiffness values

    """
    X, stresses = make_elastic_stress_random(n_samples=n_samples,
                                             elastic_modulus=elastic_modulus,
                                             poissons_ratio=poissons_ratio,
                                             size=size,
                                             macro_strain=macro_strain,
                                             grain_size=grain_size,
                                             seed=seed,
                                             volume_fraction=volume_fraction,
                                             percent_variance=percent_variance)

    return X, stresses / macro_strain



"""fMKS - functional materials knowledge system.

See https://github.com/wd15/fmks

fMKS is a functional version of [PyMKS](https://pymks.org) currently
under development. The purpose of the project is to prototype a
parallel implementation of MKS using functional programming in Python
primarily using the [Toolz](http://toolz.readthedocs.io) library.

"""

import os
import pytest
from pkg_resources import get_distribution, DistributionNotFound
from sklearn.base import TransformerMixin, BaseEstimator


def test():  # pragma: no cover
    r"""
    Run all the doctests available.
    """

    path = os.path.split(__file__)[0]
    pytest.main(
        args=[path, "--doctest-modules", "--ignore=setup.py ", "-r s", "--cov=fmks"]
    )


def get_version() -> str:
    """Get the version of the code from egg_info.

    Returns:
      the package version number
    """

    try:
        # pylint: disable=no-member
        version = get_distribution(__name__.split(".")[0]).version
    except DistributionNotFound:  # pragma: no cover
        version = "unknown, try running `python setup.py egg_info`"

    return version


__version__ = get_version()

__all__ = ["__version__", "test"]


class GenericTransformer(BaseEstimator, TransformerMixin):
    """Make a generic transformer based on a function

    >>> import numpy as np
    >>> data = np.arange(4).reshape(2, 2)
    >>> GenericTransformer(lambda x: x[:, 1:]).fit(data).transform(data).shape
    (2, 1)
    """

    def __init__(self, func):
        """Instantiate a GenericTransformer

        Function should take a multi-dimensional array and return an
        array with the same length in the sample axis (first axis).

        Args:
          func: transformer function

        """
        self.func = func

    def fit(self, *_):
        """Only necessary to make pipelines work
        """
        return self

    def transform(self, data):
        """Transform the data

        Args:
          data: the data to be transformed

        Returns:
          the transformed data
        """
        return self.func(data)



"""Functions to generate a checkerboard microstructure
"""
from dask.array import from_array
from numpy import array, reshape, indices
from toolz.curried import pipe

from ..func import curry


@curry
def _checkerboard(size, square_shape):
    """Generte a checkerboard numpy array

    `square_shape` must be a numpy array that can divide into
    `np.indices(shape)`

    Args:
      size: the size of the domain
      square_shape: the shape of each subdomain

    """
    return pipe(size, indices, lambda x: x // square_shape, lambda x: x.sum(axis=0) % 2)


def generate_checkerboard(size, square_shape=(1,)):
    """Generate a 2-phase checkerboard microstructure

    Args:
      size (tuple): the size of the domain ``(n_x, n_y)``
      square_shape (tuple): the shape of each subdomain ``(n_x, n_y)``

    Returns:
      a microstructure of shape `(1,) + shape` (extra sample axis)

    >>> print(generate_checkerboard((4,)).compute())
    [[0 1 0 1]]
    >>> print(generate_checkerboard((3, 3)).compute())
    [[[0 1 0]
      [1 0 1]
      [0 1 0]]]
    >>> print(generate_checkerboard((3, 3), (2,)).compute())
    [[[0 0 1]
      [0 0 1]
      [1 1 0]]]
    >>> print(generate_checkerboard((5, 8), (2, 3)).compute())
    [[[0 0 0 1 1 1 0 0]
      [0 0 0 1 1 1 0 0]
      [1 1 1 0 0 0 1 1]
      [1 1 1 0 0 0 1 1]
      [0 0 0 1 1 1 0 0]]]

    """
    return pipe(
        square_shape,
        array,
        lambda x: reshape(x, x.shape + (1,) * len(size)),
        _checkerboard(size),
        lambda x: from_array(x[None], chunks=-1),
    )



"""Base class for LegendreTransformer and PrimitiveTransformer
"""

from sklearn.base import TransformerMixin, BaseEstimator
import dask.array as da


class BasisTransformer(BaseEstimator, TransformerMixin):
    """Basis transformer for Sklearn pipelines

    Attributes:
      discretize: function to discretize the data
      n_state: the number of local states
      min_: the minimum local state
      max_: the maximum local state
      chunks: chunks size for state axis

    >>> import numpy as np
    >>> f = lambda *_, **__: None
    >>> BasisTransformer(f).fit().transform(np.arange(4).reshape(1, 2, 2))

    """

    # pylint: disable=too-many-arguments

    def __init__(self, discretize, n_state=2, min_=0.0, max_=1.0, chunks=None):
        """Instantiate a PrimitiveTransformer
        """
        self.discretize = discretize
        self.n_state = n_state
        self.min_ = min_
        self.max_ = max_
        self.chunks = chunks

    def transform(self, data):
        """Perform the discretization of the data

        Args:
            data: the data to discretize

        Returns:
            the discretized data
        """
        return self.discretize(
            data if hasattr(data, "chunks") else da.from_array(data, data.shape),
            n_state=self.n_state,
            min_=self.min_,
            max_=self.max_,
            chunks=self.chunks,
        )

    def fit(self, *_):
        """Only necessary to make pipelines work
        """
        return self



import pytest

def test_issue106():
    from pymks.datasets.elastic_FE_simulation import ElasticFESimulation
    import numpy as np
    L = 5
    elastic_modulus = (1, 2, 3)
    poissons_ratio = (0.3, 0.3, 0.3)
    size = (1, L, L)
    sim = ElasticFESimulation(elastic_modulus=elastic_modulus,
                              poissons_ratio=poissons_ratio)
    X = np.zeros(size, dtype=int)
    sim.run(X)
    X = np.ones(size, dtype=int)
    sim.run(X)
    X[0, 0, 0] = -1

    with pytest.raises(RuntimeError) as excinfo:
        sim.run(X)
    assert "X must be between 0 and 2." == str(excinfo.value)



"""Generate a delta microstructure for training elastic PyMKS models.

"""

from numpy import arange, transpose, array, identity
import dask.array as da

from ..func import npresize, extend, assign, sequence, curry


@curry
def _npgenerate(n_phases, shape):
    return sequence(
        arange,
        npresize((n_phases, n_phases)),
        transpose,
        extend(shape),
        assign(
            arange(n_phases)[None, :],
            ((slice(None), slice(None)) + tuple(array(shape) // 2)),
        ),
        lambda x: x[~identity(n_phases, dtype=bool)],
    )(n_phases)


@curry
def generate_delta(n_phases, shape, chunks=()):
    """Generate a delta microstructure

    A delta microstructure has a 1 at the center and 0 everywhere else
    for each phase. This is used to calibrate linear elasticity models
    that only require delta microstructures for calibration.

    Args:
      n_phases (int): number of phases
      shape (tuple): the shape of the microstructure, ``(n_x, n_y)``
      chunks (tuple): how to chunk the sample axis ``(n_chunk,)``

    Returns:
      a dask array of delta microstructures

    If `n_phases=5` for example, this requires 20 microstructures as
    each phase pairing requies 2 microstructure arrays.

    >>> arr = generate_delta(5, (3, 4), chunks=(5,))
    >>> arr.shape
    (20, 3, 4)
    >>> arr.chunks
    ((5, 5, 5, 5), (3,), (4,))
    >>> print(arr[0].compute())
    [[0 0 0 0]
     [0 0 1 0]
     [0 0 0 0]]

    `generate_delta` requires at least 2 phases

    >>> arr = generate_delta(2, (3, 3))
    >>> arr.shape
    (2, 3, 3)
    >>> print(arr[0].compute())
    [[0 0 0]
     [0 1 0]
     [0 0 0]]

    """
    return da.from_array(_npgenerate(n_phases, shape), chunks=(chunks or (-1,)) + shape)



from .filter import Filter
from scipy.linalg import lstsq
from sklearn.linear_model import LinearRegression
import numpy as np
from .fmks.func import deprecate

class MKSLocalizationModel(LinearRegression):

    """
    The `MKSLocalizationModel` fits data using the Materials Knowledge
    System in Fourier Space. The following demonstrates the viability
    of the `MKSLocalizationModel` with a simple 1D filter.

    Attributes:
        basis: Basis function used to discretize the microstucture.
        n_states: Interger value for number of local states, if a basis
            is specified, n_states indicates the order of the polynomial.
        coef_: Array of values that are the influence coefficients.

    >>> n_states = 2
    >>> n_spaces = 81
    >>> n_samples = 400

    Define a filter function.

    >>> def filter(x):
    ...     return np.where(x < 10,
    ...                     np.exp(-abs(x)) * np.cos(x * np.pi),
    ...                     np.exp(-abs(x - 20)) * np.cos((x - 20) * np.pi))

    Use the filter function to construct some coefficients.

    >>> coef_ = np.linspace(1, 0, n_states)[None,:] * filter(np.linspace(0, 20,
    ...                                                      n_spaces))[:,None]
    >>> Fcoef_ = np.fft.fft(coef_, axis=0)

    Make some test samples.

    >>> np.random.seed(2)
    >>> X = np.random.random((n_samples, n_spaces))

    Construct a response with the `Fcoef_`.

    >>> H = np.linspace(0, 1, n_states)
    >>> X_ = np.maximum(1 - abs(X[:,:,None] - H) / (H[1] - H[0]), 0)
    >>> FX = np.fft.fft(X_, axis=1)
    >>> Fy = np.sum(Fcoef_[None] * FX, axis=-1)
    >>> y = np.fft.ifft(Fy, axis=1).real

    Use the `MKSLocalizationModel` to reconstruct the coefficients

    >>> from .bases import PrimitiveBasis
    >>> prim_basis = PrimitiveBasis(n_states, [0, 1])
    >>> model = MKSLocalizationModel(basis=prim_basis)
    >>> model.fit(X, y)

    Check the result

    >>> assert np.allclose(np.fft.fftshift(coef_, axes=(0,)), model.coef_)
    """
    @deprecate
    def __init__(self, basis, n_states=None, n_jobs=1, lstsq_rcond=None):
        """
        Instantiate a MKSLocalizationModel.

        Args:
            basis (class): an instance of a bases class.
            n_states (int, optional): number of local states
            n_jobs (int, optional): number of parallel jobs to run, only used
                if pyfftw is installed.
            lstsq_rcond (float, optional): rcond argument to scipy.linalg.lstsq
                function. Defaults to 4 orders of magnitude above machine
                epsilon.

        """
        self.basis = basis
        self.n_states = n_states
        if n_states is None:
            self.n_states = basis.n_states
        self.domain = basis.domain
        self.basis._n_jobs = n_jobs
        self.lstsq_rcond = lstsq_rcond
        if self.lstsq_rcond is None:
            self.lstsq_rcond = np.finfo(float).eps*1e4

    def fit(self, X, y, size=None):
        """
        Fits the data by calculating a set of influence coefficients.

        Args:
            X (ND array): The microstructure, an `(n_samples, n_x, ...)`
                shaped array where `n_samples` is the number of samples and
                `n_x` is the spatial discretization.
            y (ND array): The response field, same shape as `X`.
            size (tuple, optional): Alters the shape of X and y during the
                calibration of the influence coefficients. If None, the size
                of the influence coefficients is the same shape as `X` and `y`.

        Example

        >>> X = np.linspace(0, 1, 4).reshape((1, 2, 2))
        >>> y = X.swapaxes(1, 2)
        >>> from .bases import PrimitiveBasis
        >>> prim_basis = PrimitiveBasis(2, [0, 1])
        >>> model = MKSLocalizationModel(basis=prim_basis)
        >>> model.fit(X, y)
        >>> assert np.allclose(model._filter._Fkernel, [[[ 0.5,  0.5],
        ...                                              [  -2,    0]],
        ...                                             [[-0.5,  0  ],
        ...                                              [  -1,  0  ]]])
        """
        self.basis = self.basis.__class__(self.n_states, self.domain)
        if size is not None:
            y = self.basis._reshape_localization_data(y, size)
            X = self.basis._reshape_feature(X, size)
        self.basis._check_shape(X.shape, y.shape)
        X_ = self.basis.discretize(X)
        FX = self.basis._fftn(X_)
        Fy = self.basis._fftn(y)
        Fkernel = np.zeros(FX.shape[1:], dtype=np.complex)
        s0 = (slice(None),)
        for ijk in np.ndindex(FX.shape[1:-1]):
            s1 = self.basis._select_slice(ijk, s0)
            Fkernel[ijk + s1] = lstsq(FX[s0 + ijk + s1], Fy[s0 + ijk],
                                      self.lstsq_rcond)[0]
        self._filter = Filter(Fkernel[None], self.basis)

    @property
    def coef_(self):
        """Returns the coefficients in real space with origin shifted to the
        center.
        """
        return self._filter._frequency_2_real(copy=True)[0]

    @coef_.setter
    def coef_(self, kernel):
        """Setter for influence coefficients.
        """
        self._filter._Fkernel = self._filter._real_2_frequency(kernel[None])
        self.basis._axes_shape = kernel.shape[:-1]

    def predict(self, X):
        """Predicts a new response from the microstructure function `X` with
        calibrated influence coefficients.

        Args:
            X (ND array): The microstructure, an `(n_samples, n_x, ...)`
                shaped array where `n_samples` is the number of samples and
                `n_x` is the spatial discretization.
        Returns:
            The predicted response field the same shape as `X`.

        Example

        >>> X = np.linspace(0, 1, 4).reshape((1, 2, 2))
        >>> y = X.swapaxes(1, 2)
        >>> from .bases import PrimitiveBasis
        >>> prim_basis = PrimitiveBasis(2, [0, 1])
        >>> model = MKSLocalizationModel(basis=prim_basis)
        >>> model.fit(X, y)
        >>> assert np.allclose(y, model.predict(X))

        The fit method must be called to calibrate the coefficients before
        the predict method can be used.

        >>> MKSModel = MKSLocalizationModel(prim_basis)
        >>> MKSModel.predict(X)
        Traceback (most recent call last):
        ...
        AttributeError: fit() method must be run before predict().
        """

        if not hasattr(self, '_filter'):
            raise AttributeError("fit() method must be run before predict().")
        _pred_shape = self.basis._pred_shape(X)
        X = self.basis._reshape_feature(X, self.basis._axes_shape)
        X_ = self.basis.discretize(X)
        return self._filter.convolve(X_).reshape(_pred_shape).real

    def resize_coeff(self, size):
        """Scale the size of the coefficients and pad with zeros.

        Args:
            size (tuple): The new size of the influence coefficients.

        Returns:
            The resized influence coefficients to size.

        Example

        Let's first instantitate a model and fabricate some
        coefficients.

        >>> from pymks.bases import PrimitiveBasis
        >>> prim_basis = PrimitiveBasis(n_states=1)
        >>> prim_basis._axes = np.array([1, 2])
        >>> prim_basis._axes_shape = (5, 4)
        >>> model = MKSLocalizationModel(prim_basis)
        >>> coef_ = np.arange(20).reshape((1, 5, 4, 1))
        >>> coef_ = np.concatenate((coef_, np.ones_like(coef_)), axis=-1)
        >>> coef_ = np.fft.ifftshift(coef_, axes=(1, 2))
        >>> model._filter = Filter(np.fft.rfftn(coef_, axes=(1, 2)),
        ...                        prim_basis)

        The coefficients can be reshaped by passing the new shape that
        coefficients should have.

        >>> model.resize_coeff((10, 7))
        >>> assert np.allclose(model.coef_[... ,0],
        ...                    [[0, 0, 0, 0, 0, 0, 0],
        ...                     [0, 0, 0, 0, 0, 0, 0],
        ...                     [0, 0, 0, 0, 0, 0, 0],
        ...                     [0, 0, 0, 1, 2, 3, 0],
        ...                     [0, 0, 4, 5, 6, 7, 0],
        ...                     [0, 0, 8, 9,10,11, 0],
        ...                     [0, 0,12,13,14,15, 0],
        ...                     [0, 0,16,17,18,19, 0],
        ...                     [0, 0, 0, 0, 0, 0, 0],
        ...                     [0, 0, 0, 0, 0, 0, 0]])
        """
        self._filter.resize(size)

    def _test(self):
        """Tests

        >>> n_states = 10
        >>> np.random.seed(3)
        >>> X = np.random.random((2, 5, 3))
        >>> from .bases import PrimitiveBasis
        >>> prim_basis = PrimitiveBasis(n_states, [0, 1])
        >>> X_ = prim_basis.discretize(X)
        >>> H = np.linspace(0, 1, n_states)
        >>> Xtest = np.sum(X_ * H[None,None,None,:], axis=-1)
        >>> assert np.allclose(X, Xtest)

        Use Legendre polynomials basis for the microstructure function
        and take the Fourier transform.

        >>> from .bases import LegendreBasis
        >>> np.random.seed(3)
        >>> X = np.random.random((1, 3, 3))
        >>> leg_basis = LegendreBasis(2, [0, 1])
        >>> model = MKSLocalizationModel(basis=leg_basis)
        >>> X_ = leg_basis.discretize(X)
        >>> FX = np.fft.fftn(X_, axes=(1, 2))
        >>>
        >>> FXtest = np.array([[[[4.50000000+0.j, -0.79735949+0.],
        ...                      [0.00000000+0.j, -1.00887157-1.48005289j],
        ...                      [0.00000000+0.j, -1.00887157+1.48005289j]],
        ...                     [[0.00000000+0.j, 0.62300683-4.97732233j],
        ...                      [0.00000000+0.j, 1.09318216+0.10131035j],
        ...                      [0.00000000+0.j, 0.37713401+1.87334545j]],
        ...                     [[0.00000000+0.j, 0.62300683+4.97732233j],
        ...                      [0.00000000+0.j, 0.37713401-1.87334545j],
        ...                      [0.00000000+0.j, 1.09318216-0.10131035j]]]])
        >>> assert np.allclose(FX, FXtest)
        """
        pass



"""Top level functions for the Graspi graph descriptors for materials
science.

"""
# flake8: noqa
import pandas
import numpy as np
import pytest
from toolz.curried import keymap
from sklearn.base import TransformerMixin, BaseEstimator

from .func import fmap, curry, pipe, apply_dataframe_func, make_da_return


@curry
def graph_descriptors_np(data, delta_x=1.0, periodic_boundary=True):
    """Numpy only version of graph_descriptors function

    Args:
      data (array): array of phases ``(n_samples, n_x, n_y)``, values must be 0 or 1
      delta_x (float): pixel size
      periodic_boundary (bool): whether the boundaries are periodic

    Returns:
      A Pandas data frame with samples along rows and descriptors
      along columns

    """
    # pylint: enable=line-too-long
    columns = keymap(
        lambda x: x.encode("UTF-8"),
        dict(
            STAT_n="n_vertices",
            STAT_e="n_edges",
            STAT_n_D="n_phase0",
            STAT_n_A="n_phase1",
            STAT_CC_D="n_phase0_connect",
            STAT_CC_A="n_phase1_connect",
            STAT_CC_D_An="n_phase0_connect_top",
            STAT_CC_A_Ca="n_phase1_connect_bottom",
            ABS_wf_D="w_frac_phase0",
            ABS_f_D="frac_phase0",
            DISS_wf10_D="w_frac_phase0_10_dist",
            DISS_f10_D="fraction_phase0_10_dist",
            DISS_f2_D="fraction_phase0_2_dist",
            CT_f_conn_D="frac_useful",
            CT_f_e_conn="inter_frac_bottom_and_top",
            CT_f_conn_D_An="frac_phase0_top",
            CT_f_conn_A_Ca="frac_phase1_bottom",
            CT_e_conn="n_inter_paths",
            CT_e_D_An="n_phase0_inter_top",
            CT_e_A_Ca="n_phase1_inter_bottom",
            CT_f_D_tort1="frac_phase0_rising",
            CT_f_A_tort1="frac_phase1_rising",
            CT_n_D_adj_An="n_phase0_connect_anode",
            CT_n_A_adj_Ca="n_phase1_connect_cathode",
        ),
    )

    return pipe(
        data,
        fmap(
            graph_descriptors_sample(
                delta_x=delta_x, periodic_boundary=periodic_boundary
            )
        ),
        list,
        # debug('check fields in graspi'),
        pandas.DataFrame,
        lambda x: x.rename(columns=columns),
        lambda x: x.apply(
            lambda x: np.rint(x).astype(int) if x.name[:2] == "n_" else x
        ),
    )


@make_da_return
@curry
def graph_descriptors(data, delta_x=1.0, periodic_boundary=True):
    # pylint: disable=line-too-long
    """Compute graph descriptors for multiple samples

    Args:
      data: array of phases (n_samples, n_x, n_y), values must be 0 or 1
      delta_x: pixel size
      periodic_boundary: whether the boundaries are periodic

    Returns:
      A Pandas data frame with samples along rows and descriptors
      along columns

    Compute graph descriptors for multiple samples using the GraSPI
    sub-package. See the `installation instructions
    <rst/README.html#graspi>`_ to install PyMKS with
    GraSPI enabled.

    GraSPI is focused on characterizing photovoltaic devices and so
    the descriptors must be understood in this context. Future
    releases will have more generic descriptors. See `Wodo et
    al. <https://arxiv.org/pdf/1106.3536.pdf>`_ for more details.
    Note that the current implementation only works for two phase
    data.

    This function returns a Pandas Dataframe with the descriptors as
    columns and samples in rows. In the context of a photovoltaic
    device the top of the domain (y-direction) represents an anode and
    the bottom of the domain represents a cathode. Phase 0 represents
    donor materials while phase 1 represents acceptor material. Many
    of these descriptors characterizes the morphology in terms of hole
    electron pair generation and transport leading to device charge
    extraction.

    To use `graph_descriptors` as part of a Sklearn pipeline, see
    :class:`~pymks.GraphDescriptors`.

    The column descriptors are as follows.

    ========================= ===========
    Column Name               Description
    ========================= ===========
    n_vertices                The number of vertices in the constructed graph. Should be equal to the number of pixels.
    n_edges                   The number of edges in the constructed graph.
    n_phase{i}                The number of vertices for phase {i}.
    n_phase{i}_connect        The number of connected components for phase {i}.
    n_phase{i}_connect_top    The number of connected components for phase {i} with the top of the domain in y-direction.
    n_phase{i}_connect_bottom The number of connected components for phase {i} with the top of the domain in y-direction.
    w_frac_phase{i}           Weighted fraction of phase {i} vertices.
    frac_phase{i}             Fraction of phase {i} vertices.
    w_frac_phase{i}_{j}_dist  Weighted fraction of phase {i} vertices within j nodes from an interface.
    frac_phase{i}_{j}_dist    Fraction of phase {i} vertices within {j} nodes from an interface.
    frac_useful               Fraction of useful vertices connected the top or bottom of the domain.
    inter_frac_bottom_top     Fraction of interface with complementary paths to bottom or top of the domain.
    frac_phase{i}_top         Fraction of phase {i} interface vertices with path to top.
    frac_phase{i}_bottom      Fraction of phase {i} interface vertices with path to bottom.
    n_inter_paths             Number of interface edges with complementary paths.
    n_phase{i}_inter_top      Number of phase {i} interface vertices with path to top
    n_phase{i}_inter_bottom   Number of phase {i} interface vertices with path to bottom
    frac_phase{i}_rising      Fraction of phase {i} with rising paths
    ========================= ===========


    Example, with 3 x (3, 3) arrays

    Read in the expected data.

    >>> from io import StringIO
    >>> expected = pandas.read_csv(StringIO('''
    ... n_vertices,n_edges,n_phase0,n_phase1,n_phase0_connect,n_phase1_connect,n_phase0_connect_top,n_phase1_connect_bottom,w_frac_phase0,frac_phase0,w_frac_phase0_10_dist,fraction_phase0_10_dist,inter_frac_bottom_and_top,frac_phase0_top,frac_phase1_bottom,n_inter_paths,n_phase0_inter_top,n_phase1_inter_bottom,frac_phase0_rising,frac_phase1_rising,n_phase0_connect_anode,n_phase1_connect_cathode
    ... 9,7,3,6,2,1,1,1,0.3256601095199585,0.3333333432674408,0.9624541997909546,1.0,0.4285714328289032,0.3333333432674408,1.0,3,1,6,1.0,0.6666666865348816,2,2
    ... 9,6,3,6,1,1,1,1,0.3267437815666199,0.3333333432674408,0.9624541997909546,1.0,1.0,1.0,1.0,6,3,6,1.0,1.0,2,3
    ... 9,6,6,3,2,1,1,0,0.6534984707832336,0.6666666865348816,0.9624541997909546,1.0,0.0,0.5,0.0,0,3,0,1.0,0.0,4,1
    ... '''))

    Construct the 3 samples each with 3x3 voxels

    >>> data = np.array([[[0, 1, 0],
    ...                   [0, 1, 1],
    ...                   [1, 1, 1]],
    ...                  [[1, 1, 1],
    ...                   [0, 0, 0],
    ...                   [1, 1, 1]],
    ...                  [[0, 1, 0],
    ...                   [0, 1, 0],
    ...                   [0, 1, 0]]])
    >>> actual = graph_descriptors(data)

    ``graph_descriptors`` returns a data frame.

    >>> actual
       n_vertices  n_edges  ...  n_phase0_connect_anode  n_phase1_connect_cathode
    0           9        7  ...                       2                         2
    1           9        6  ...                       2                         3
    2           9        6  ...                       4                         1
    <BLANKLINE>
    [3 rows x 22 columns]


    Check that the actual values are equal to the expected values.

    >>> assert np.allclose(actual, expected)

    Works with Dask arrays as well. When using Dask a Dask dataframe
    will be returned.

    >>> import dask.array as da
    >>> out = graph_descriptors(da.from_array(data, chunks=(2, 3, 3)))
    >>> out.get_partition(0).compute()
       n_vertices  n_edges  ...  n_phase0_connect_anode  n_phase1_connect_cathode
    0           9        7  ...                       2                         2
    1           9        6  ...                       2                         3
    <BLANKLINE>
    [2 rows x 22 columns]

    On examining the data for this simple test case there are a few
    obvious checks. Each sample has 9 vertices since there are 9
    pixels in each sample.

    >>> actual.n_vertices
    0    9
    1    9
    2    9
    Name: n_vertices, dtype: int64

    Notice that the first and third sample have two phase 1 regions
    connected to either the top or bottom of the domain while the
    second sample has only 1 region.

    >>> actual.n_phase1_connect
    0    1
    1    1
    2    1
    Name: n_phase1_connect, dtype: int64

    All paths are blocked for the first and second samples from
    reaching the top from the bottom surface. The third sample has 6
    interface edges that connect the top and bottom.

    >>> actual.n_inter_paths
    0    3
    1    6
    2    0
    Name: n_inter_paths, dtype: int64

    """
    return apply_dataframe_func(
        graph_descriptors_np(delta_x=delta_x, periodic_boundary=periodic_boundary), data
    )


@curry
def graph_descriptors_sample(data, delta_x=1.0, periodic_boundary=True):
    """Calculate graspi graph descriptors for a single array
    """
    graspi = pytest.importorskip("graspi")

    compute = lambda x: graspi.compute_descriptors(
        x,
        *(data.shape + (3 - len(data.shape)) * (1,)),
        pixelS=delta_x,
        if_per=periodic_boundary
    )

    return pipe(
        data,
        lambda x: x.astype(np.int32).flatten(order="F"),
        compute,
        fmap(lambda x: x[::-1]),
        dict,
    )


class GraphDescriptors(BaseEstimator, TransformerMixin):
    """Calculate GraphDescriptors as part of a Sklearn pipeline

    Wraps the :func:`~pymks.graph_descriptors` function

    Test

    >>> data = np.array([[[0, 1, 0],
    ...                   [0, 1, 1],
    ...                   [1, 1, 1]],
    ...                  [[1, 1, 1],
    ...                   [0, 0, 0],
    ...                   [1, 1, 1]],
    ...                  [[0, 1, 0],
    ...                   [0, 1, 0],
    ...                   [0, 1, 0]]])
    >>> actual = GraphDescriptors().fit(data).transform(data)
    >>> actual.shape
    (3, 22)

    See the :func:`~pymks.graph_descriptors` function for more
    complete documentation.

    """

    def __init__(self, delta_x=1.0, periodic_boundary=True):
        """Instantiate a GraphDescriptors transformer

        Args:
          delta_x: pixel size
          periodic_boundary: whether the boundaries are periodic
          columns: subset of columns to include

        """
        self.delta_x = delta_x
        self.periodic_boundary = periodic_boundary

    def transform(self, data):
        """Transform the data

        Args:
          data: the data to be transformed

        Returns:
          the graph descriptors dataframe
        """
        return graph_descriptors(
            data, delta_x=self.delta_x, periodic_boundary=self.periodic_boundary
        )

    def fit(self, *_):
        """Only necessary to make pipelines work
        """
        return self


class GenericTransformer(BaseEstimator, TransformerMixin):
    """Make a generic transformer based on a function

    >>> data = np.arange(4).reshape(2, 2)
    >>> GenericTransformer(lambda x: x[:, 1:]).fit(data).transform(data).shape
    (2, 1)
    """

    def __init__(self, func):
        """Instantiate a GenericTransformer

        Function should take a multi-dimensional array and return an
        array with the same length in the sample axis (first axis).

        Args:
          func: transformer function

        """
        self.func = func

    def fit(self, *_):
        """Only necessary to make pipelines work
        """
        return self

    def transform(self, data):
        """Transform the data

        Args:
          data: the data to be transformed

        Returns:
          the transformed data
        """
        return self.func(data)



import numpy as np
from pymks.bases import GSHBasis


def test_gsh_no_symmetry():
    """this test checks that a particular gsh basis function with no
    symmetry is being evaluated properly"""

    X = np.array([[0.1, 0.2, 0.3],
                  [6.5, 2.3, 3.4]])
    gsh_basis = GSHBasis(n_states=[1])

    assert(np.allclose(np.squeeze(gsh_basis.discretize(X)), q_no_symm(X)))


def test_gsh_hex():
    """this test checks that a particular gsh basis function for hexagonal
    symmetry is being evaluated properly"""

    X = np.array([[0.1, 0.2, 0.3],
                  [6.5, 2.3, 3.4]])
    gsh_basis = GSHBasis(n_states=[1], domain='hexagonal')

    assert(np.allclose(np.squeeze(gsh_basis.discretize(X)), q_hex(X)))


def test_symmetry_check_hex():
    """this test is designed to check that the hexagonal gsh functions
    for two symmetrically equivalent orientations output the same gsh
    coefficients"""

    X1 = np.array([[30, 70, 45]])*np.pi/180.
    X2 = np.array([[30+180, 180-70, 2*60-45]])*np.pi/180.
    gsh_basis = GSHBasis(n_states=np.arange(0, 100, 5), domain='hexagonal')

    assert(np.allclose(gsh_basis.discretize(X1), gsh_basis.discretize(X2)))


def q_hex(x):
    phi1 = x[:, 0]
    phi = x[:, 1]
    t913 = np.sin(phi)
    x_GSH = -((0.5e1 / 0.4e1) * np.exp((-2*1j) * phi1) *
              np.sqrt(0.6e1) * t913 ** 2)
    return x_GSH


def q_no_symm(x):
    phi1 = x[:, 0]
    phi = x[:, 1]
    phi2 = x[:, 2]
    x_GSH = ((0.3e1 / 0.2e1) * (0.1e1 + np.cos(phi)) *
             np.exp((-1*1j) * (phi1 + phi2)))
    return x_GSH


if __name__ == '__main__':
    test_gsh_no_symmetry()
    test_gsh_hex()
    test_symmetry_check_hex()



import numpy as np


def test_frequency_2_real_and_back():
    from pymks.filter import Filter
    from pymks import PrimitiveBasis

    X = np.zeros((1, 3, 2, 2))
    X[0, 0, 0] = np.arange(1, 3) * 9
    p_basis = PrimitiveBasis(2)
    p_basis._axes = (1, 2)
    p_basis._axes_shape = (3, 3)
    X_result = np.ones((1, 3, 3, 2)) * np.arange(1, 3)[None, None, None, :]
    filter_ = Filter(X, p_basis)
    assert np.allclose(filter_._frequency_2_real(copy=True), X_result)
    assert np.allclose(filter_._real_2_frequency(X_result), X)

if __name__ == '__main__':
    test_frequency_2_real_and_back()



"""Functional helper functions for fmks.
"""

from functools import wraps

import numpy as np
import dask.array as da
from dask import delayed
import dask.dataframe as ddf
import toolz.curried
from toolz.curried import iterate, compose, pipe, get, flip, identity
from toolz.curried import map as fmap
import deprecated


def curry(func):
    """Curry a function, but keep the docstring

    Args:
      func: a function to curry

    Returns:
      a curried function

    >>> def add(a, b):
    ...     '''Add stuff'''
    ...     return a + b
    >>> curry(add)(1)(2)
    3
    >>> print(curry(add).__doc__)
    Add stuff
    """
    return wraps(func)(toolz.curried.curry(func))


@curry
def array_from_tuple(data, shape, dtype):
    """Create an array from a list of slices and values.

    Args:
      data: list of slices and values used to create a new array
      shape: the shape of the new array
      dtype: the type of the new array


    Returns:
      the new populated array

    >>> array_from_tuple(
    ...     (((slice(None), slice(2)), 3.), ((2, 2), -1)),
    ...     (3, 3),
    ...     int
    ... )
    array([[ 3,  3,  0],
           [ 3,  3,  0],
           [ 3,  3, -1]])
    """
    arr = np.zeros(shape, dtype=dtype)
    for slice_, value in data:
        arr[slice_] = value
    return arr


@curry
def iterate_times(func, times, value):
    """Iterate a function over a value.

    Args:
      func: the function to iterate
      times: the number of times to iterate
      value: the value to iterate over

    Returns:
      an update value

    >>> def inc(value):
    ...     return value + 1

    >>> iterate_times(inc, 0, 1)
    1

    >>> iterate_times(inc, 3, 1)
    4
    """
    iter_ = iterate(func, value)
    for _ in range(times):
        next(iter_)
    return next(iter_)


@curry
def map_blocks(func, data, chunks=None, dtype=None):
    """Curried version of Dask's map_blocks

    Args:
      func: the function to map
      data: a Dask array
      chunks: chunks for new array if reshaped

    Returns:
      a new Dask array

    >>> f = map_blocks(lambda x: x + 1)
    >>> f(da.arange(4, chunks=(2,)))
    dask.array<lambda, shape=(4,), dtype=int64, chunksize=(2,), chunktype=numpy.ndarray>
    """
    return da.map_blocks(func, data, chunks=chunks, dtype=dtype)


allclose = curry(np.allclose)  # pylint: disable=invalid-name

fft = curry(np.fft.fft)  # pylint: disable=invalid-name

ifft = curry(np.fft.ifft)  # pylint: disable=invalid-name

fftn = curry(np.fft.fftn)  # pylint: disable=invalid-name

ifftn = curry(np.fft.ifftn)  # pylint: disable=invalid-name

fftshift = curry(np.fft.fftshift)  # pylint: disable=invalid-name

ifftshift = curry(np.fft.ifftshift)  # pylint: disable=invalid-name

daifftn = curry(da.fft.ifftn)  # pylint: disable=invalid-name

dafftn = curry(da.fft.fftn)  # pylint: disable=invalid-name

dafft = curry(da.fft.fft)  # pylint: disable=invalid-name

daifft = curry(da.fft.ifft)  # pylint: disable=invalid-name

dafftshift = curry(da.fft.fftshift)  # pylint: disable=invalid-name

daifftshift = curry(da.fft.ifftshift)  # pylint: disable=invalid-name

daconj = curry(da.conj)  # pylint: disable=invalid-name

dapad = curry(da.pad)  # pylint: disable=invalid-name


def rcompose(*args):
    """Compose functions in order

    Args:
      args: the functions to compose

    Returns:
      composed functions

    >>> assert rcompose(lambda x: x + 1, lambda x: x * 2)(3) == 8
    """
    return compose(*args[::-1])


sequence = rcompose  # pylint: disable=invalid-name


def apply_dict_func(func, data, shape_dict):

    """Apply a function that returns a dictionary of arrays to a Dask
    array in parallel.

    e.g.

    >>> data = da.from_array(np.arange(36).reshape(4, 3, 3), chunks=(2, 3, 3))
    >>> def func(arr):
    ...     return dict(
    ...         a=np.resize(
    ...             arr,
    ...             (arr.shape[0],) + (arr.shape[1] + 1,) + (arr.shape[2] + 1,)
    ...         ),
    ...         b=np.resize(arr, (arr.shape[:3]) + (1,))
    ...     )
    >>> out = apply_dict_func(func, data, dict(a=(4, 4, 4), b=(4, 3, 3, 1)))
    >>> print(out['a'].chunks)
    ((2, 2), (4,), (4,))
    >>> print(out['b'].shape)
    (4, 3, 3, 1)
    >>> print(out['a'].compute().shape)
    (4, 4, 4)

    Args:
      func: the function to apply to the Dask array
      data: the Dask array to call the function with
      shape_dict: a dictionary of shapes, the keys are the keys
        returned by the funcion and the shapes correspond to the
        shapes that are output from the function

    """

    @curry
    def from_delayed(key, shape, delayed_func):
        return da.from_delayed(
            delayed_func, dtype=float, shape=(shape[0],) + shape_dict[key][1:]
        )

    def concat(key):
        return pipe(
            lambda x: func(np.array(x)),
            delayed,
            lambda x: fmap(lambda y: (y.shape, x(y)), data.blocks),
            fmap(lambda x: (x[0], get(key, x[1]))),
            fmap(lambda x: from_delayed(key, x[0], x[1])),
            list,
            lambda x: da.concatenate(x, axis=0),
        )

    return pipe(shape_dict.keys(), fmap(lambda x: (x, concat(x))), dict)


@curry
def debug(stmt, data):  # pragma: no cover
    """Helpful debug function
    """
    print(stmt)
    print(data)
    return data


def flatten(data):
    """Flatten data along all but the first axis

    Args:
      data: data to flatten

    Returns:
      the flattened data

    >>> data = np.arange(18).reshape((2, 3, 3))
    >>> flatten(data).shape
    (2, 9)
    """
    return data.reshape(data.shape[0], -1)


@curry
def rechunk(chunks, data):
    """An agnostic rechunk for numpy or dask

    Required as from_array no longer accepts dask arrays.

    Args:
      chunks: the new chunk shape
      data: either a numpy or dask array


    Returns:
      a rechunked dask array

    >>> rechunk((1, 5), np.arange(10).reshape((2, 5))).chunks
    ((1, 1), (5,))

    """
    if isinstance(data, np.ndarray):
        rechunk_ = da.from_array
    else:
        rechunk_ = da.rechunk
    return rechunk_(data, chunks=chunks)


def make_da(func):
    """Decorator to allow functions that only take Dask arrays to take
    Numpy arrays.

    Args:
      func: the function to be decorated

    Returns:
      the decorated function

    >>> @make_da
    ... def my_func(arr):
    ...     return arr + 1

    >>> my_func(np.array([1, 1]))
    dask.array<add, shape=(2,), dtype=int64, chunksize=(2,), chunktype=numpy.ndarray>

    """

    def wrapper(arr, *args, **kwargs):
        if isinstance(arr, np.ndarray):
            chunks = arr.shape
        else:
            chunks = arr.chunks
        return func(rechunk(chunks, arr), *args, **kwargs)

    return wraps(func)(wrapper)


def make_da_return(func):
    """Decorator to allow functions that only take Dask arrays to take
    Numpy arrays, but then return a Numpy array if given a Numpy array

    Args:
      func: the function to be decorated

    Returns:
      the decorated function

    >>> @make_da_return
    ... def my_func(darr):
    ...     return da.from_array(
    ...         np.array(darr) + 1,
    ...         chunks=darr.chunks
    ...     )

    >>> my_func(np.arange(10))
    array([ 1,  2,  3,  4,  5,  6,  7,  8,  9, 10])

    >>> my_func(da.arange(10, chunks=(2,)))
    dask.array<array, shape=(10,), dtype=int64, chunksize=(2,), chunktype=numpy.ndarray>

    """

    def wrapper(arr, *args, **kwargs):
        isnd = lambda x: isinstance(x, np.ndarray)
        compute = lambda x: x.compute() if hasattr(x, "compute") else x
        return pipe(
            arr,
            rechunk(arr.shape if isnd(arr) else arr.chunks),
            lambda x: func(x, *args, **kwargs),
            (compute if isnd(arr) else identity),
        )

    return wraps(func)(wrapper)


def make_da_return2(func):
    """Decorator to allow functions that take two array arguments to only
    take Dask arrays to take Numpy arrays, but then return a Numpy
    array if given a Numpy array

    Args:
      func: the function to be decorated

    Returns:
      the decorated function

    >>> @make_da_return2
    ... def my_func(darr1, darr2):
    ...     return da.from_array(
    ...         np.array(darr1) + 1,
    ...         chunks=darr1.chunks
    ...     ) + da.from_array(
    ...         np.array(darr2) + 1,
    ...         chunks=darr2.chunks
    ...     )

    >>> my_func(np.arange(10), np.arange(10))
    array([ 2,  4,  6,  8, 10, 12, 14, 16, 18, 20])

    >>> my_func(da.arange(10, chunks=(2,)), da.arange(10, chunks=(2,)))
    dask.array<add, shape=(10,), dtype=int64, chunksize=(2,), chunktype=numpy.ndarray>

    If either of the arrays are numpy then a numpy array is returned

    >>> my_func(da.arange(10, chunks=(2,)), np.arange(10))
    array([ 2,  4,  6,  8, 10, 12, 14, 16, 18, 20])

    >>> my_func(np.arange(10), da.arange(10, chunks=(2,)))
    array([ 2,  4,  6,  8, 10, 12, 14, 16, 18, 20])

    """

    def wrapper(arr1, arr2, *args, **kwargs):
        return make_da_return(curry(make_da_return(func))(arr1))(arr2, *args, **kwargs)

    return wraps(func)(wrapper)


@curry
def extend(shape, arr):
    """Extend an array by adding new axes with shape of shape argument.

    The values from the existing axes are repeated in the new
    axes. The is achieved using repeated uses of np.repeat followed by
    np.reshape.

    Args:
      shape: the new shape to extend by
      arr: the array to extend

    Returns:
      a new extended array

    >>> a = np.arange(6).reshape((2, 3))
    >>> extend((4,), a).shape
    (2, 3, 4)
    >>> print(extend((2, 3), a))
    [[[[0 0 0]
       [0 0 0]]
    <BLANKLINE>
      [[1 1 1]
       [1 1 1]]
    <BLANKLINE>
      [[2 2 2]
       [2 2 2]]]
    <BLANKLINE>
    <BLANKLINE>
     [[[3 3 3]
       [3 3 3]]
    <BLANKLINE>
      [[4 4 4]
       [4 4 4]]
    <BLANKLINE>
      [[5 5 5]
       [5 5 5]]]]

    """
    extend_ = curry(lambda s, x: np.repeat(x, s, axis=-1).reshape(x.shape + (s,)))
    fextend = sequence(fmap(extend_), list, lambda x: sequence(*x))
    return fextend(shape)(arr)


@curry
def assign(value, index, arr):
    """Fake functional numpy assignment

    Just to make things easier for function pipelines

    Args:
      value: the value to assing
      index: the index
      arr: the array

    Returns:
      Not a new array, but the same array updated

    >>> a = np.arange(6).reshape((2, 3))
    >>> assign(1, (slice(None), 2), a)
    array([[0, 1, 1],
           [3, 4, 1]])

    """
    arr[index] = value
    return arr


npresize = curry(flip(np.resize))  # pylint: disable=invalid-name


@curry
def zero_pad(arr, shape, chunks):
    """Zero pad an array with zeros

    Args:
      arr: the array to pad
      shape: the shape of the new array
      chunks: how to rechunk the new array

    Returns:
      the new padded version of the array

    >>> print(
    ...     zero_pad(
    ...         np.arange(4).reshape([1, 2, 2, 1]),
    ...         (1, 4, 5, 1),
    ...         None
    ...     )[0,...,0].compute()
    ... )
    [[0 0 0 0 0]
     [0 0 0 1 0]
     [0 0 2 3 0]
     [0 0 0 0 0]]
    >>> print(zero_pad(np.arange(4).reshape([2, 2]), (4, 5), None).compute())
    [[0 0 0 0 0]
     [0 0 0 1 0]
     [0 0 2 3 0]
     [0 0 0 0 0]]
    >>> zero_pad(zero_pad(np.arange(4).reshape([2, 2]), (4, 5, 1), None))
    Traceback (most recent call last):
    ...
    RuntimeError: length of shape is incorrect
    >>> zero_pad(zero_pad(np.arange(4).reshape([2, 2]), (1, 2), None))
    Traceback (most recent call last):
    ...
    RuntimeError: resize shape is too small

    >>> arr = da.from_array(np.arange(4).reshape((2, 2)), chunks=(2, 1))
    >>> out = zero_pad(arr, (4, 3), (-1, 1))
    >>> out.shape
    (4, 3)
    >>> out.chunks
    ((4,), (1, 1, 1))
    """
    if len(shape) != len(arr.shape):
        raise RuntimeError("length of shape is incorrect")

    if not np.all(shape >= arr.shape):
        raise RuntimeError("resize shape is too small")

    return pipe(
        np.array(shape) - np.array(arr.shape),
        lambda x: np.concatenate(
            ((x - (x // 2))[..., None], (x // 2)[..., None]), axis=1
        ),
        fmap(tuple),
        tuple,
        lambda x: da.pad(arr, x, "constant", constant_values=0),
        lambda x: da.rechunk(x, chunks=chunks or x.shape),
    )


@curry
def star(func, args):
    """Allow function to take piped sequence as arguments.

    Args:
      func: any function
      args: a sequence of arguments to the function

    Returns:
      evaluated function

    >>> star(lambda x, y: x * y)([2, 3])
    6
    """
    return func(*args)


def deprecate(func, reason=None):
    """Deprecation warning for PyMKS.

    >>> @deprecate
    ... def my_func():
    ...     return

    >>> my_func()

    """

    @deprecated.deprecated(version=0.4, reason=reason or "Use pymks.fmks instead.")
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def dist_mesh(shape):
    """Calculate a mesh of distances

    Assume dx=1 and the center pixel is shifted to the left for odd
    shaped axes.

    Args:
      shape: the shape of the domain

    Returns:
      an array of distances from the center

    >>> assert np.allclose(
    ...    dist_mesh((3, 2, 4)),
    ...    np.sqrt(np.array(
    ...     [[[6, 3, 2, 3],
    ...       [5, 2, 1, 2]],
    ...      [[5, 2, 1, 2],
    ...       [4, 1, 0, 1]],
    ...      [[6, 3, 2, 3],
    ...       [5, 2, 1, 2]]]
    ...     ))
    ... )

    >>> dist_mesh((3, 5, 4)).shape
    (3, 5, 4)

    """

    center = lambda x: np.reshape(np.array(x) // 2, (len(x),) + (1,) * len(x))

    return sequence(
        fmap(lambda x: np.linspace(0, x - 1, x)),
        lambda x: np.array(np.meshgrid(*x, indexing="ij")),
        lambda x: x - center(x.shape[1:]),
        lambda x: np.linalg.norm(x, axis=0),
    )(shape)


def sort_array(arr):
    """Functional sort

    Args:
      arr: array to sort

    Returns:
      tuple of the sorted array and the index ordering

    >>> sort_array(np.array([2, 1, 3]))
    (array([1, 2, 3]), array([1, 0, 2]))

    """
    return sequence(np.argsort, lambda x: (arr[x], x))(arr)


def apply_dataframe_func(func, data):
    """Daskerize a function that takes an array and returns a dataframe

    >>> import pandas

    >>> def func(x):
    ...     return pandas.DataFrame(x)

    >>> x = np.random.random((10, 4))

    >>> df = apply_dataframe_func(func, da.from_array(x, chunks=(2, 4)))
    >>> df.get_partition(0).compute().shape
    (2, 4)

    """
    return pipe(data.blocks, fmap(delayed(func)), list, ddf.from_delayed)



"""Test the localization model.
"""

from sklearn.pipeline import make_pipeline
import numpy as np
import dask.array as da
from pymks.fmks.bases.primitive import discretize, redundancy
from pymks.fmks.localization import fit
from pymks.fmks.bases.primitive import PrimitiveTransformer
from pymks.fmks.localization import LocalizationRegressor
from pymks.fmks.data.delta import generate_delta
from pymks.fmks.data import solve_fe


def _get_x():
    return da.from_array(np.linspace(0, 1, 4).reshape((1, 2, 2)), chunks=(1, 2, 2))


def test():
    """Very simple example.
    """
    assert np.allclose(
        fit(
            _get_x(),
            _get_x().swapaxes(1, 2),
            discretize(n_state=2),
            redundancy_func=redundancy,
        ),
        [[[0.5, 0.5], [-2, 0]], [[-0.5, 0], [-1, 0]]],
    )


def test_setting_kernel():
    """Test resetting the coeffs after coeff resize.
    """

    x_data = generate_delta(n_phases=2, shape=(21, 21)).persist()

    y_data = solve_fe(
        x_data, elastic_modulus=(100, 130), poissons_ratio=(0.3, 0.3), macro_strain=0.01
    )["strain"][..., 0].persist()

    model = make_pipeline(PrimitiveTransformer(n_state=2), LocalizationRegressor())

    shape = (30, 30)
    fcoeff = model.fit(x_data, y_data).steps[1][1].coeff
    assert np.allclose(model.steps[1][1].coeff_resize(shape).coeff.shape[:-1], shape)
    model.steps[1][1].coeff = fcoeff
    assert np.allclose(model.predict(x_data), y_data, atol=1e-4)



import pytest
import numpy as np
from pymks.datasets import make_microstructure


def test_size_and_grain_size_failure():
    with pytest.raises(RuntimeError):
        make_microstructure(n_samples=1, size=(7, 7), grain_size=(8, 1))


def test_volume_fraction_failure():
    with pytest.raises(RuntimeError):
        make_microstructure(n_samples=1, volume_fraction=(0.3, 0.6))


def test_volume_fraction_with_n_phases_failure():
    with pytest.raises(RuntimeError):
        make_microstructure(n_samples=1, size=(7, 7), n_phases=3,
                            volume_fraction=(0.5, 0.5))


def test_percent_variance_exceeds_limit_failure():
    with pytest.raises(RuntimeError):
        make_microstructure(n_samples=1, size=(7, 7), n_phases=3,
                            volume_fraction=(0.3, 0.3, 0.4), percent_variance=0.5)


def test_volume_fraction():
    X = make_microstructure(n_samples=1, n_phases=3,
                            volume_fraction=(0.3, 0.2, 0.5))
    assert np.allclose(np.sum(X == 1) / float(X.size), 0.2, rtol=1e-4)
    assert np.allclose(np.sum(X == 2) / float(X.size), 0.5, atol=1e-4)


def test_percent_variance():
    volume_fraction = (0.3, 0.2, 0.5)
    percent_variance = .2
    X = make_microstructure(n_samples=1, n_phases=3,
                            volume_fraction=(0.3, 0.2, 0.5),
                            percent_variance=percent_variance)
    vf_1 = np.sum(X == 1) / float(X.size)
    vf_2 = np.sum(X == 2) / float(X.size)
    assert vf_1 > volume_fraction[1] - percent_variance
    assert vf_1 < volume_fraction[1] + percent_variance
    assert vf_2 > volume_fraction[2] - percent_variance
    assert vf_2 < volume_fraction[2] + percent_variance

def test_3D_volume_fraction():
    X = make_microstructure(n_samples=1, n_phases=3, size=(21, 21, 21),
                           volume_fraction=(0.3, 0.1, 0.6))
    assert np.allclose(np.sum(X == 0) / float(X.size), 0.3, rtol=1e-3)
    assert np.allclose(np.sum(X == 1) / float(X.size), 0.1, rtol=1e-3)

def test_3D_percent_variance():
    volume_fraction = (0.7, 0.1, 0.2)
    percent_variance = .2
    X = make_microstructure(n_samples=1, n_phases=3, size=(21, 21, 21),
                            volume_fraction=volume_fraction,
                            percent_variance=percent_variance)
    vf_1 = np.sum(X == 1) / float(X.size)
    vf_2 = np.sum(X == 2) / float(X.size)
    assert vf_1 > volume_fraction[1] - percent_variance
    assert vf_1 < volume_fraction[1] + percent_variance
    assert vf_2 > volume_fraction[2] - percent_variance
    assert vf_2 < volume_fraction[2] + percent_variance

if __name__ == '__main__':
    test_3D_volume_fraction()
    test_3D_percent_variance()



r"""Primitive basis for microstructure discretization

Example.

Discretize the microstructure function into `n_states` local states
such that:

.. math::

   \frac{1}{\Delta x} \int_{H} \int_{s} \Lambda(h - l)
   m(h, x) dx dh = m[l, s]

where :math:`\Lambda` is the primitive basis (also called hat basis)
function

.. math::

   \Lambda (h - l) = max \Bigg (1-\Bigg |\frac{h(L - 1)}{H} -
   \frac{Hl}{L-1} \Bigg|, 0\Bigg)

A microstructure function discretized with this basis is subject to
the following constraint

.. math::

   \sum_{l=0}^L m[l, s] = 1

which is equivalent of saying that every location is filled with some
configuration of local states.

Here is an example with 3 discrete local states in a microstructure.

>>> x_data = np.array([[[1, 1, 0],
...                     [1, 0 ,2],
...                     [0, 1, 0]]])
>>> assert(x_data.shape == (1, 3, 3))

>>> x_test = np.array([[[[0, 1, 0],
...                      [0, 1, 0],
...                      [1, 0, 0]],
...                     [[0, 1, 0],
...                      [1, 0, 0],
...                      [0, 0, 1]],
...                     [[1, 0, 0],
...                      [0, 1, 0],
...                      [1, 0, 0]]]])

The when a microstructure is discretized, the different local states are
mapped into local state space, which results in an array of shape
`(n_samples, n_x, n_y, n_states)`, where `n_states=3` in this case.
For example, if a cell has a label of 2, its local state will be
`[0, 0, 1]`. The local state can only have values of 0 or 1.

>>> from toolz import pipe
>>> assert pipe(
...     x_data,
...     discretize(n_state=3, max_=2),
...     curry(np.allclose)(x_test))

"""

import dask.array as da
import numpy as np
from ..func import curry
from .basis_transformer import BasisTransformer


def discretize_nomax(data, states):
    """Helper function for primitive discretization.

    Args:
      data: the data to discretize
      states: a sequence of local states

    Returns:
      the unnormalized discetized microstructure

    Example:

    >>> discretize_nomax(np.array([[0, 1]]), np.array([0., 0.5, 1.0]))
    array([[[ 1.,  0., -1.],
            [-1.,  0.,  1.]]])

    >>> discretize_nomax(da.linspace(0, 1, 9, chunks=(3,)),
    ...             da.linspace(0, 1, 6, chunks=(2,))).shape
    (9, 6)

    """
    return 1 - (abs(data[..., None] - states)) / (states[1] - states[0])


@curry
def discretize(x_data, n_state=2, min_=0.0, max_=1.0, chunks=None):
    """Primitive discretization of a microstructure.

    Args:
      x_data: the data to discretize
      n_state: the number of local states
      min_: the minimum local state
      max_: the maximum local state
      chunks: chunks size for state axis

    Returns:
      the discretized microstructure

    >>> discretize(da.random.random((12, 9), chunks=(3, 9)),
    ...            3,
    ...            chunks=1).chunks
    ((3, 3, 3, 3), (9,), (1, 1, 1))

    >>> discretize(np.array([[0, 1], [0.5, 0.5]]), 3, chunks=1).chunks
    ((2,), (2,), (1, 1, 1))

    >>> assert np.allclose(
    ...     discretize(
    ...         np.array([[0, 1], [0.5, 0.5]]),
    ...         3,
    ...         chunks=1
    ...     ).compute(),
    ...     [[[1, 0, 0], [0, 0, 1]], [[0, 1, 0], [0, 1, 0]]]
    ... )
    """
    return da.maximum(
        discretize_nomax(
            da.clip(x_data, min_, max_),
            da.linspace(min_, max_, n_state, chunks=(chunks or n_state,)),
        ),
        0,
    )


def redundancy(ijk):
    """Used in localization to remove redundant slices

    Args:
      ijk: the current index

    Returns:
      the redundant slice, (slice(-1),) when no redundancies
    """
    if np.all(np.array(ijk) == 0):
        return (slice(None),)

    return (slice(-1),)


class PrimitiveTransformer(BasisTransformer):
    """Primitive transformer for Sklearn pipelines

    >>> from toolz import pipe
    >>> assert pipe(
    ...     PrimitiveTransformer(),
    ...     lambda x: x.fit(None, None),
    ...     lambda x: x.transform(np.array([[0, 0.5, 1]])).compute(),
    ...     lambda x: np.allclose(x,
    ...         [[[1. , 0. ],
    ...           [0.5, 0.5],
    ...           [0. , 1. ]]])
    ... )
    """

    def __init__(self, n_state=2, min_=0.0, max_=1.0, chunks=None):
        """Instantiate a PrimitiveTransformer

        Args:
           n_state: the number of local states
           min_: the minimum local state
           max_: the maximum local state
           chunks: chunks size for state axis

        """
        super().__init__(
            discretize, n_state=n_state, min_=min_, max_=max_, chunks=chunks
        )



import numpy as np
from ..fmks.func import deprecate


class _AbstractMicrostructureBasis(object):
    @deprecate
    def __init__(self, n_states=2, domain=None):
        """
        Instantiate a `Basis`

        Args:
            n_states (int, list): The number of local states, or an array of
                local states to be used.
            domain (list, optional): indicate the range of expected values for
                the microstructure, default is [0, n_states - 1].
        """
        if hasattr(n_states, '__len__') and len(n_states) > 0:
            self.n_states = n_states
        else:
            self.n_states = np.arange(n_states)
        if domain is None:
            domain = [0, max(self.n_states)]
        self.domain = domain
        self._n_jobs = 1

    def check(self, X):
        if (np.min(X) < self.domain[0]) or (np.max(X) > self.domain[1]):
            raise RuntimeError("X must be within the specified domain")

    def discretize(self, X):
        raise NotImplementedError

    def _check_shape(self, X_shape, y_shape):
        if not len(y_shape) > 1:
            raise RuntimeError("The shape of y is incorrect.")
        if y_shape != X_shape:
            raise RuntimeError("X and y must be the same shape.")

    def _pred_shape(self, X):
        """
        Function to describe the expected output shape of a given
        microstructure X.
        """
        return X.shape

    def _select_slice(self, ijk, s0):
        """
        Helper method used to calibrate influence coefficients from in
        mks_localization_model to account for redundancies from linearly
        dependent local states.
        """
        return s0

    def _reshape_feature(self, X, size):
        """
        Helper function used to check the shape of the microstructure,
        and change to appropriate shape.

        Args:
            X: The microstructure, an `(n_samples, n_x, ...)` shaped array
                where `n_samples` is the number of samples and `n_x` is the
                patial discretization.
            size: the new size of the array

        Returns:
            microstructure with shape (n_samples, size)
        """
        return X.reshape((X.shape[0],) + size)

    def _reshape_localization_data(self, y, size):
        """
        Helper function used to check the shape of the microstructure,
        and change to appropriate shape.

        Args:
            y: The localization fields, an `(n_samples, n_x, ...)` shaped array
                where `n_samples` is the number of samples and `n_x` is thes
                patial discretization.
            size: the new size of the array

        Returns:
            Localization fields with shape (n_samples, size)
        """
        return y.reshape((y.shape[0],) + size)

    def _select_axes(self, X):
        self._axes = np.arange(X.ndim - 1) + 1
        self._axes_shape = X[0].shape



"""Test functions in func.py
"""

import numpy as np
import dask.array as da

from pymks.fmks.func import make_da


def test_make_da():
    """Test bug in make_da.

    make_da was not preserving the chunks of dask arrays
    """
    arr = np.arange(24).reshape((4, 6))
    darr = da.from_array(arr, chunks=(2, 6))

    def my_func(darr):
        return darr

    new_func = make_da(my_func)

    assert new_func(darr).chunks == ((2, 2), (6,))



import numpy as np


def test_nonperiodic_autocorrelation():
    '''
    test nonperiodic autocorrelation for spatial statistics
    '''
    from pymks import DiscreteIndicatorBasis
    from pymks.stats import autocorrelate
    X = np.array([[[1, 0, 1, 1],
                   [1, 0, 1, 1],
                   [0, 0, 0, 0],
                   [0, 0, 0, 0],
                   [0, 0, 0, 0]]])
    basis = DiscreteIndicatorBasis(n_states=2)
    X_auto = autocorrelate(X, basis)

    X_result = np.array([[[0,       0,       0,       0],
                          [1. / 8, 1. / 12, 3. / 16, 1. / 12],
                          [0.2, 2. / 15,     0.3, 2. / 15],
                          [1. / 8, 1. / 12, 3. / 16, 1. / 12],
                          [0,       0,       0,       0]]])

    assert(np.allclose(X_result, X_auto[..., 1]))


def test_periodic_autocorrelation():
    '''
    test periodic autocorrelation for spatial statistics
    '''
    from pymks import DiscreteIndicatorBasis
    from pymks.stats import autocorrelate
    X = np.array([[[1, 0, 1, 1],
                   [1, 0, 1, 1],
                   [0, 0, 0, 0],
                   [0, 0, 0, 0],
                   [0, 0, 0, 0]]])
    basis = DiscreteIndicatorBasis(n_states=2)
    X_auto = autocorrelate(X, basis, periodic_axes=(0, 1))

    X_result = np.array([[[0,   0,    0,   0],
                          [0.1, 0.1, 0.15, 0.1],
                          [0.2, 0.2,  0.3, 0.2],
                          [0.1, 0.1, 0.15, 0.1],
                          [0,   0,    0,   0]]])

    assert(np.allclose(X_result, X_auto[..., 1]))


def test_nonperiodic_crosscorrelation():
    '''
    test nonperiodic crosscorrelation
    '''
    from pymks import DiscreteIndicatorBasis
    from pymks.stats import crosscorrelate
    X = np.array([[[1, 0, 1, 1],
                   [1, 0, 1, 1],
                   [0, 0, 0, 0],
                   [0, 0, 0, 0],
                   [0, 0, 0, 0]]])
    basis = DiscreteIndicatorBasis(n_states=2)
    X_cross = crosscorrelate(X, basis)

    X_result = np.array([[[1 / 3., 4 / 9., 0.5,  4 / 9., ],
                          [1 / 8., 0.25, 3 / 16., 0.25],
                          [0., 2 / 15.,  0., 2 / 15.],
                          [0., 1 / 12., 0, 1 / 12.],
                          [0,   0,    0,   0]]])
    assert(np.allclose(X_result, X_cross[..., 0]))


def test_periodic_crosscorrelation():
    '''
    test periodic crosscorrelation
    '''
    from pymks import DiscreteIndicatorBasis
    from pymks.stats import crosscorrelate
    X = np.array([[[1, 0, 1, 1],
                   [1, 0, 1, 1],
                   [0, 0, 0, 0],
                   [0, 0, 0, 0],
                   [0, 0, 0, 0]]])
    basis = DiscreteIndicatorBasis(n_states=2)
    X_cross = crosscorrelate(X, basis, periodic_axes=(0, 1))

    X_result = np.array([[[0.3, 0.3, 0.3,  0.3],
                          [0.2, 0.2, 0.15, 0.2],
                          [0.1, 0.1,  0., 0.1],
                          [0.2, 0.2, 0.15, 0.2],
                          [0.3, 0.3, 0.3,  0.3]]])

    assert(np.allclose(X_result, X_cross[..., 0]))


def test_nonperiodic_correlate():
    '''
    test corrleate for non-periodic microstructures
    '''
    from pymks import DiscreteIndicatorBasis
    from pymks.stats import correlate

    X = np.array([[[0, 0, 1, 0],
                   [0, 0, 1, 0],
                   [0, 0, 1, 0],
                   [0, 0, 0, 0],
                   [0, 0, 1, 0]],
                  [[0, 1, 0, 0],
                   [0, 1, 0, 0],
                   [0, 1, 0, 0],
                   [0, 0, 0, 0],
                   [0, 1, 0, 0]]])
    basis = DiscreteIndicatorBasis(n_states=2)
    X_corr = correlate(X, basis)
    X_result = [[2 / 3.,  4 / 9.,  0.75,  4 / 9.],
                [5 / 8.,   0.5,  0.75,   0.5],
                [0.6,  7 / 15.,   0.8,  7 / 15.],
                [5 / 8.,   0.5,  0.75,   0.5],
                [0.5,  4 / 9.,  0.75,  4 / 9.]]
    assert(np.allclose(X_result, X_corr[0, ..., 0]))


def test_periodic_correlate():
    '''
    test corrleate for periodic microstructures
    '''
    from pymks import DiscreteIndicatorBasis
    from pymks.stats import correlate

    X = np.array([[[0, 0, 1, 0],
                   [0, 0, 1, 0],
                   [0, 0, 1, 0],
                   [0, 0, 0, 0],
                   [0, 0, 1, 0]],
                  [[0, 1, 0, 0],
                   [0, 1, 0, 0],
                   [0, 1, 0, 0],
                   [0, 0, 0, 0],
                   [0, 1, 0, 0]]])
    basis = DiscreteIndicatorBasis(n_states=2)
    X_corr = correlate(X, basis, periodic_axes=(0, 1))
    X_result = [[0.6,  0.6,  0.75,  0.6],
                [0.6,  0.6,  0.75,  0.6],
                [0.6,  0.6,   0.8,  0.6],
                [0.6,  0.6,  0.75,  0.6],
                [0.6,  0.6,  0.75,  0.6]]
    assert(np.allclose(X_result, X_corr[0, ..., 0]))


def test_periodic_mask():
    '''
    test uncertainty masks for periodic axes.
    '''
    from pymks import DiscreteIndicatorBasis
    from pymks.stats import autocorrelate
    from pymks.datasets import make_checkerboard_microstructure

    X = make_checkerboard_microstructure(1, 3)
    basis = DiscreteIndicatorBasis(n_states=2)
    mask = np.ones((X.shape))
    mask[0, 0, 0] = 0
    X_auto_periodic_mask = autocorrelate(X, basis, periodic_axes=[0, 1],
                                         confidence_index=mask)
    X_result_0 = np.array([[[1 / 7., 1 / 7., 3 / 7.],
                          [1 / 7., 0.5, 1 / 7.],
                          [3 / 7., 1 / 7., 1 / 7.]]])
    X_result_1 = np.array([[[2 / 7., 1 / 7., 2 / 7.],
                          [1 / 7., 0.5, 1 / 7.],
                          [2 / 7., 1 / 7., 2 / 7.]]])
    X_result = np.concatenate((X_result_0[..., None],
                               X_result_1[..., None]), axis=-1)
    assert np.allclose(X_auto_periodic_mask, X_result)


def test_nonperiodic_mask():
    '''
    test uncertainty masks for nonperiodic axes.
    '''
    from pymks import DiscreteIndicatorBasis
    from pymks.stats import autocorrelate
    from pymks.datasets import make_checkerboard_microstructure

    X = make_checkerboard_microstructure(1, 3)
    basis = DiscreteIndicatorBasis(n_states=2)
    mask = np.ones((X.shape))
    mask[0, 0, 0] = 0
    X_auto_nonperiodic_mask = autocorrelate(X, basis, confidence_index=mask)
    X_result_0 = np.array([[[1 / 3., 0, 0.5],
                          [0, 0.5, 0.],
                          [0.5, 0, 1 / 3.]]])
    X_result_1 = np.array([[[2 / 3., 0, 0.5],
                          [0, 0.5, 0.],
                          [0.5, 0, 2 / 3.]]])
    X_result = np.concatenate((X_result_0[..., None],
                               X_result_1[..., None]), axis=-1)
    assert np.allclose(X_auto_nonperiodic_mask, np.concatenate(X_result))


def test_mixperdic_mask():
    from pymks import DiscreteIndicatorBasis
    from pymks.stats import autocorrelate
    from pymks.datasets import make_checkerboard_microstructure

    X = make_checkerboard_microstructure(1, 3)
    basis = DiscreteIndicatorBasis(n_states=2)
    mask = np.ones((X.shape))
    mask[0, 0, 0] = 0
    X_auto_mixperiodic_mask = autocorrelate(X, basis, periodic_axes=[0],
                                            confidence_index=mask)
    X_result_0 = np.array([[[1 / 5., 1 / 7., 2 / 5.],
                          [0, 0.5, 0],
                          [2 / 5., 1 / 7., 1 / 5.]]])
    X_result_1 = np.array([[[2 / 5., 1 / 7., 2 / 5.],
                          [0, 0.5, 0.],
                          [2 / 5., 1 / 7., 2 / 5.]]])
    X_result = np.concatenate((X_result_0[..., None],
                               X_result_1[..., None]), axis=-1)
    assert np.allclose(X_auto_mixperiodic_mask, np.concatenate(X_result))


def test_mask_two_samples():
    from pymks import DiscreteIndicatorBasis
    from pymks.stats import correlate
    from pymks.datasets import make_microstructure

    X = make_microstructure(n_samples=2, n_phases=2, size=(3, 3),
                            grain_size=(2, 2), seed=99)
    basis = DiscreteIndicatorBasis(n_states=2)
    mask = np.ones(X.shape)
    mask[:, 0, 0] = 0.
    X_corr = correlate(X, basis, confidence_index=mask)
    X_result = np.array([[[[1 / 3., 1 / 3., 1 / 3.],
                           [1 / 5., 1 / 5., 1 / 5.],
                           [1 / 4., 1 / 4., 0]],
                          [[1 / 5., 1 / 5., 2 / 5.],
                           [1 / 2., 1 / 2., 0],
                           [1 / 5., 1 / 5., 1 / 5.]],
                          [[1 / 4., 1 / 4., 1 / 2.],
                           [1 / 5., 1 / 5., 2 / 5.],
                           [1 / 3., 1 / 3., 0]]],
                         [[[0., 0., 1 / 3.],
                           [2 / 5., 3 / 5., 0.],
                           [0., 0., 1 / 2.]],
                          [[0., 0., 2 / 5.],
                           [3 / 8., 5 / 8., 0],
                           [0., 0., 3 / 5.]],
                          [[0., 0., 1 / 2.],
                           [2 / 5., 3 / 5., 0.],
                           [0., 0., 2 / 3.]]]])
    assert np.allclose(X_corr, X_result)


def test_normalization_rfftn():
    """Test normalization with rfftn
    """
    from pymks import PrimitiveBasis
    from pymks.stats import _normalize
    prim_basis = PrimitiveBasis()
    Nx = Ny = 5
    X_ = np.zeros((1, Nx, Ny, 1))
    prim_basis._axes = np.arange(X_.ndim - 2) + 1
    prim_basis._axes_shape = (2 * Nx, 2 * Ny)
    norm = _normalize(X_.shape, prim_basis, None)
    assert norm.shape == (1, Nx, Ny, 1)
    assert np.allclose(norm[0, Nx // 2, Ny // 2, 0], 25)


def test_normalization_fftn():
    """Test normalization with fftn
    """
    from pymks.bases import FourierBasis
    from pymks.stats import _normalize
    f_basis = FourierBasis()
    Nx = Ny = 5
    X_ = np.zeros((1, Nx, Ny, 1))
    f_basis._axes = np.arange(X_.ndim - 2) + 1
    f_basis._axes_shape = (2 * Nx, 2 * Ny)
    norm = _normalize(X_.shape, f_basis, None)
    assert norm.shape == (1, Nx, Ny, 1)
    assert np.allclose(norm[0, Nx // 2, Ny // 2, 0], 25)


def test_gsh_basis_normalization():
    from pymks.bases import GSHBasis
    from pymks.stats import _normalize
    gsh_basis = GSHBasis()
    Nx = Ny = 5
    X_ = np.zeros((1, Nx, Ny, 1))
    gsh_basis._axes = np.arange(X_.ndim - 2) + 1
    gsh_basis._axes_shape = (2 * Nx, 2 * Ny)
    norm = _normalize(X_.shape, gsh_basis, None)
    assert norm.shape == (1, Nx, Ny, 1)
    assert np.allclose(norm[0, Nx // 2, Ny // 2, 0], 25)


def test_stats_in_parallel():
    from pymks.bases import PrimitiveBasis
    from pymks.stats import correlate
    from pymks.datasets import make_microstructure
    p_basis = PrimitiveBasis(5)
    X = make_microstructure(n_samples=5, n_phases=3)
    X_corr_actual = correlate(X, p_basis)
    for i in range(1, 4):
        X_corr_test = correlate(X, p_basis, n_jobs=i)
        assert np.allclose(X_corr_actual, X_corr_test)


def test_autocorrelate_with_specific_correlations():
    from pymks.stats import autocorrelate
    from pymks import PrimitiveBasis
    X = np.array([[[1, 0, 1, 1],
                   [1, 0, 1, 1],
                   [0, 0, 2, 0],
                   [0, 0, 0, 0],
                   [0, 0, 0, 0]]])
    autocorrelations = [(0, 0), (2, 2)]
    p_basis = PrimitiveBasis(n_states=3)
    X_auto = autocorrelate(X, p_basis, autocorrelations=autocorrelations)
    X_result_0 = np.array([[2 / 3., 1 / 3., 5 / 12., 4 / 9.],
                           [5 / 8., 5 / 12., 9 / 16., 1 / 2.],
                           [1 / 2., 7 / 15., 13 / 20., 7 / 15.],
                           [3 / 8., 1 / 2., 9 / 16., 5 / 12.],
                           [1 / 6., 4 / 9., 5 / 12., 1 / 3.]])
    assert np.allclose(X_auto[0, ..., 0], X_result_0)
    X_result_1 = np.array([[0., 0., 0., 0.],
                           [0., 0., 0., 0.],
                           [0., 0., 0.05, 0.],
                           [0., 0., 0., 0.],
                           [0., 0., 0., 0.]])
    assert np.allclose(X_auto[0, ..., 1], X_result_1)


def test_crosscorrelate_with_specific_correlations():
    from pymks.stats import crosscorrelate
    from pymks import PrimitiveBasis
    X = np.array([[[0, 0, 0, 0],
                   [0, 1, 0, 0],
                   [0, 0, 2, 0],
                   [0, 0, 0, 0],
                   [0, 0, 0, 0]]])
    crosscorrelations = [(1, 2)]
    p_basis = PrimitiveBasis(n_states=3)
    X_cross = crosscorrelate(X, p_basis, crosscorrelations=crosscorrelations)
    X_result = np.array([[0., 0., 0., 0.],
                         [0., 0., 0., 0.],
                         [0., 0., 0., 0.],
                         [0., 0., 0., 1 / 12.],
                         [0., 0., 0., 0.]])
    assert np.allclose(X_cross[0, ..., 0], X_result)


if __name__ == '__main__':
    test_normalization_rfftn()



from ..filter import Filter
from scipy.ndimage.fourier import fourier_gaussian
from .base_microstructure_generator import _BaseMicrostructureGenerator
import numpy as np


class MicrostructureGenerator(_BaseMicrostructureGenerator):
    """
    Generates n_samples number of a periodic random microstructures
    with domain size equal to size and with n_phases number of
    phases. The optional grain_size argument controls the size and
    shape of the grains.

    >>> n_samples, n_phases = 1, 2
    >>> size = (4, 4)
    >>> generator = MicrostructureGenerator(n_samples, size,
    ...                                      n_phases, seed=10)
    >>> X = generator.generate()
    >>> X_test = np.array([[[1, 0, 1, 1],
    ...                     [1, 0, 0, 1],
    ...                     [0, 0, 1, 1],
    ...                     [0, 1, 1, 1]]])
    >>> assert(np.allclose(X, X_test))

    """

    def generate(self):
        """
        Generates a microstructure of dimensions of self.size and grains
        with dimensions self.grain_size.

        Returns:
          periodic microstructure
        """
        if len(self.size) != len(self.grain_size):
            raise RuntimeError("Dimensions of size and grain_size are"
                               " not equal.")

        X = np.random.random((self.n_samples,) + self.size)
        _gaussian_size = np.around(self.grain_size).astype(int)
        gaussian = fourier_gaussian(np.ones(_gaussian_size),
                                    np.ones(len(self.size)))
        filter_ = Filter(self._fftn(gaussian[None, ..., None]), self)
        filter_.resize(self.size)
        X_blur = filter_.convolve(X[..., None]).real
        return self._assign_phases(X_blur).astype(int)

    def _assign_phases(self, X_blur):
        """
        Takes in blurred array and assigns phase values.

        Args:
          X_blur: random field that has be blurred by a convolution.
        Returns:
          microstructure with assigned phases
        """
        if self.volume_fraction is None:
            epsilon = 1e-5
            X0, X1 = np.min(X_blur), np.max(X_blur)
            Xphases = float(self.n_phases) * ((X_blur - X0) / (X1 - X0) *
                                              (1. - epsilon) + epsilon)
            X_phases = np.floor(Xphases - epsilon)
        else:
            v_cum = np.cumsum(self.volume_fraction[:-1])
            X_sort = np.sort(X_blur.reshape((X_blur.shape[0], -1)), axis=1)
            seg_shape = (len(X_sort), len(v_cum))
            per_diff = (2 * np.random.random(seg_shape) -
                        1) * np.array(self.percent_variance)
            seg_ind = np.floor((v_cum + per_diff) *
                               X_sort.shape[1]).astype(int)
            seg_values = np.concatenate([x[list(i)][None]
                                         for i, x in zip(seg_ind, X_sort)])
            new_axes = [None for i in X_blur[0].shape]
            seg_slices = [slice(None)] + new_axes + [slice(None)]
            X_bool = X_blur[..., None] > seg_values[tuple(seg_slices)]
            X_phases = np.sum(X_bool, axis=-1)
        return X_phases



#  tests for pymks-0.4.1-py37h03978a9_1 (this is a generated file);
print('===== testing package: pymks-0.4.1-py37h03978a9_1 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import pymks

pymks.test()
#  --- run_test.py (end) ---

print('===== pymks-0.4.1-py37h03978a9_1 OK =====');
print("import: 'pymks'")
import pymks

